"""
ConsistentTreesArbor io classes and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
import os

from ytree.data_structures.io import \
    DataFile, \
    TreeFieldIO
from ytree.frontends.rockstar.io import \
    RockstarDataFile

class ConsistentTreesDataFile(DataFile):
    def open(self):
        self.fh = open(self.filename, "r")

class ConsistentTreesTreeFieldIO(TreeFieldIO):
    def _read_fields(self, root_node, fields, dtypes=None,
                     root_only=False):
        """
        Read fields from disk for a single tree.
        """

        data_file = self.arbor.data_files[root_node._fi]

        if dtypes is None:
            dtypes = {}
        my_dtypes = self._determine_dtypes(
            fields, override_dict=dtypes)

        close = False
        if data_file.fh is None:
            close = True
            data_file.open()
        fh = data_file.fh
        fh.seek(root_node._si)
        if root_only:
            data = [fh.readline()]
        else:
            data = fh.read(
                root_node._ei -
                root_node._si).split("\n")
            if len(data[-1]) == 0:
                data.pop()
        if close:
            data_file.close()

        nhalos = len(data)
        field_data = {}
        fi = self.arbor.field_info
        for field in fields:
            field_data[field] = np.empty(nhalos, dtype=my_dtypes[field])

        for i, datum in enumerate(data):
            ldata = datum.strip().split()
            if len(ldata) == 0: continue
            for field in fields:
                dtype = my_dtypes[field]
                field_data[field][i] = dtype(ldata[fi[field]["column"]])

        self._apply_units(fields, field_data)

        return field_data

class ConsistentTreesHlistDataFile(RockstarDataFile):
    def _parse_header(self):
        super()._parse_header()

        prefix = os.path.join(os.path.dirname(self.filename), "hlist_")
        suffix = ".list"
        self.scale_factor = self.arbor._get_file_index(
            self.filename, prefix, suffix)



"""
AHFArbor class and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from collections import defaultdict
import glob
import os
import re

from ytree.data_structures.arbor import \
    CatalogArbor
from ytree.frontends.ahf.fields import \
    AHFFieldInfo, \
    AHFCRMFieldInfo
from ytree.frontends.ahf.io import \
    AHFDataFile, \
    AHFCRMDataFile
from ytree.frontends.ahf.misc import \
    parse_AHF_file
from unyt.unit_registry import \
    UnitRegistry
from ytree.utilities.io import \
    f_text_block

class AHFArbor(CatalogArbor):
    """
    Arbor for Amiga Halo Finder data without a CRMratio2 file.

    In this instance, we will use the mtree files to assemble the tree.
    """

    _ahf_prefix = None
    _mtree_prefix = None
    _data_suffix = ".AHF_halos"
    _mtree_suffix = ".AHF_mtree"
    _par_suffix = ".parameter"
    _crm_prefix = "MergerTree_"
    _crm_suffix = "-CRMratio2"
    crm_filename = None
    _frequent_crm_midfixes = (".txt", "")
    _field_info_class = AHFFieldInfo
    _data_file_class = AHFDataFile

    def __init__(self, filename,
                 log_filename=None,
                 parameter_filename=None,
                 crm_filename=None,
                 hubble_constant=1.0, box_size=None,
                 omega_matter=None, omega_lambda=None,
                 name_config=None):
        self.unit_registry = UnitRegistry()
        self.hubble_constant = hubble_constant
        self.omega_matter = omega_matter
        self.omega_lambda = omega_lambda
        self._box_size_user = box_size

        self.log_filename = log_filename
        self.parameter_filename = parameter_filename
        self.crm_filename = crm_filename

        if name_config is None:
            name_config = {}
        self._set_naming_conventions(name_config)

        self._file_pattern = re.compile(
            rf"(^.+[^0-9a-zA-Z]+)(\d+).*{self._par_suffix}$")
        super().__init__(filename)

    def _set_naming_conventions(self, config):
        """
        Set some filename conventions.
        """

        for k, v in config.items():
            if not re.match(r"\w+_(?:prefix|suffix)", k):
                raise ValueError(
                    f"name_config entry must end in either prefix or suffix: \"{k}\"")
            if not hasattr(self, f"_{k}"):
                raise ValueError(
                    f"name_config entry not associated with a known attribute: \"{k}\"")

            setattr(self, f"_{k}", v)

    @classmethod
    def _is_crm_file(cls, filename):
        """
        Checking if this has the proper crm prefix and suffix.
        """

        return os.path.basename(filename).startswith(cls._crm_prefix) and \
          filename.endswith(cls._crm_suffix)

    def _guess_crm_filename(self, filename):
        """
        Take an educated guess at the crm filename.

        Return None if our guess does not exist.
        """

        if self.crm_filename is not None:
            if os.path.exists(self.crm_filename):
                return self.crm_filename
            else:
                return None

        if AHFArbor._is_crm_file(filename):
            if os.path.exists(filename):
                return filename
            else:
                return None

        # Searching for <keyword>.something.<suffix>
        res = re.search(rf"([^\.]+)\.[^\.]+{self._par_suffix}$", filename)
        if not res:
            return None

        filekey = res.groups()[0]
        ddir = os.path.dirname(filekey)

        for midfix in self._frequent_crm_midfixes:
            bname = os.path.basename(filekey) + midfix
            fname = os.path.join(ddir, f"{self._crm_prefix}{bname}{self._crm_suffix}")
            if os.path.exists(fname):
                return fname

        return None

    @classmethod
    def _is_parameter_file(cls, filename):
        """
        Check if file has the proper suffix.
        """

        return filename.endswith(cls._par_suffix)

    def _parse_parameter_file(self):
        df = AHFDataFile(self.parameter_filename, self)

        pars = {"simu.omega0": "omega_matter",
                "simu.lambda0": "omega_lambda",
                "simu.boxsize": "box_size"}

        if self.log_filename is None:
            fns = glob.glob(df.filekey + "*.log")
            if fns:
                log_filename = fns[0]
            else:
                log_filename = None
        else:
            log_filename = self.log_filename

        if log_filename is not None and os.path.exists(log_filename):
            vals = parse_AHF_file(log_filename, pars, sep=":")
            for attr in ["omega_matter",
                         "omega_lambda"]:
                setattr(self, attr, vals.get(attr))
            if "box_size" in vals:
                self.box_size = self.quan(vals["box_size"], "Mpc/h")

        if self._box_size_user is not None:
            self.box_size = self.quan(self._box_size_user, "Mpc/h")

        # fields from from the .AHF_halos files
        f = open(f"{df.data_filekey}{self._data_suffix}")
        line = f.readline()
        f.close()

        fields = [key[:key.rfind("(")]
                  for key in line[1:].strip().split()]
        fi = dict([(field, {"column": i, "file": "halos"})
                   for i, field in enumerate(fields)])

        # the scale factor comes from the catalog file header
        fields.append("redshift")
        fi["redshift"] = {"file": "header", "units": ""}

        # the descendent ids come from the .AHF_mtree files
        fields.append("desc_id")
        fi["desc_id"] = {"file": "mtree", "units": ""}

        self.field_list = fields
        self.field_info.update(fi)

    _fprefix = None
    @property
    def _prefix(self):
        if self._fprefix is None:
            # Match a patten of any characters, followed by some sort of
            # separator (e.g., "." or "_"), then a number, and eventually
            # the suffix.
            reg = self._file_pattern.search(self.parameter_filename)
            self._fprefix = reg.groups()[0]
        return self._fprefix

    def _get_data_files(self):
        """
        Get all *.parameter files and sort them in reverse order.
        """
        my_files = glob.glob(f"{self._prefix}*{self._par_suffix}")
        # sort by catalog number
        my_files.sort(key=self._get_file_index)
        self.data_files = \
          [self._data_file_class(f, self) for f in my_files]

        # Set the mtree file for file i to that of i-1, since
        # AHF thinks in terms of progenitors and not descendents.
        for i, data_file in enumerate(self.data_files[:-1]):
            data_file.mtree_filename = \
              self.data_files[i+1].mtree_filename
        self.data_files[-1].mtree_filename = None
        self.data_files.reverse()

    def _get_file_index(self, f):
        reg = self._file_pattern.search(f)
        if not reg:
            raise RuntimeError(
                f"Could not locate index within file: {f}.")
        return int(reg.groups()[1])

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        File must end in .parameter.
        """
        fn = args[0]
        if not AHFArbor._is_parameter_file(fn):
            return False

        if "crm_filename" in kwargs:
            return False

        if self._guess_crm_filename(self, fn) is not None:
            return False

        return True

class AHFCRMArbor(AHFArbor):
    """
    Arbor for AHF data that includes a CRM file.

    The CRM file contains all the halo links, so we don't have to
    assemble them from the mtree files.
    """

    _has_uids = True
    _field_info_class = AHFCRMFieldInfo
    _data_file_class = AHFCRMDataFile

    def _set_paths(self, filename):
        """
        Get both a CRM file and a parameter file.

        Raise an exception if we can't get both.
        """

        super()._set_paths(filename)

        if self.crm_filename is None:
            self.crm_filename = self._guess_crm_filename(filename)
        else:
            if not os.path.exists(self.crm_filename):
                raise RuntimeError(
                    "Specified crm_filename does not exist: ",
                    self.crm_filename)

        if self.crm_filename is None:
            raise RuntimeError(
                f"crm_filename is None for {type(self)}. This shouldn't be.")

        if self.parameter_filename is None:
            self.parameter_filename = self._guess_parameter_filename(filename)
        else:
            if not os.path.exists(self.parameter_filename):
                raise RuntimeError(
                    "Specified parameter_filename does not exist: ",
                    self.parameter_filename)

        if self.parameter_filename is None:
            raise RuntimeError(
                f"parameter_filename is None for {type(self)}. This shouldn't be.")

    def _set_parameter_filename(self, filename):
        """
        We have a whole thing here...
        """

        pass

    def _guess_parameter_filename(self, filename):
        """
        Guess parameter filename from the crm filename.
        """

        if self.parameter_filename is not None:
            if os.path.exists(self.parameter_filename):
                return self.parameter_filename
            else:
                return None

        basename = os.path.basename(filename)
        filekey = basename[len(self._crm_prefix):-len(self._crm_suffix)]

        # try a couple guesses at naming conventions,
        # but don't work too hard.
        for midfix in self._frequent_crm_midfixes:
            pfns = glob.glob(
                os.path.join(self.directory, filekey[:-len(midfix)]) +
                f"*{self._par_suffix}")
            if pfns:
                break

        # just look for all files with the right suffix and hope for the best
        if not pfns:
            pfns = glob.glob(f"*{self._par_suffix}")

        if not pfns:
            raise RuntimeError(
                f"Could not find any files ending in {self._par_suffix} "
                "from which to get parameters. Put some of those in here.")

        try:
            pfns.sort(key=self._get_file_index)
            return pfns[-1]

        except Exception:
            return None

    def _plant_trees(self):
        if self.is_planted:
            return

        self._compute_links()
        super()._plant_trees()

    def _compute_links(self):
        """
        Read the CRMratio2 file and hand out a dictionary of
        uid: desc_uid for each data file.
        """

        links = defaultdict(dict)

        f = open(self.crm_filename, mode="r")
        for i in range(3):
            f.readline()

        for line, loc in f_text_block(f, pbar_string="Computing links"):
            if line.startswith("END"):
                break

            online = line.split()
            thing = online[0]
            if len(online) == 2:
                my_descid = int(thing)
                continue

            my_id = int(thing)
            cid = int(thing[:-12])
            links[cid][my_id] = my_descid
        f.close()

        for df in self.data_files:
            df._links = links[df._catalog_index]

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        Filename must end in .parameter or match the CRM naming
        convention.
        """
        fn = args[0]
        # if it is a crm file, then we are good to go
        if AHFCRMArbor._is_crm_file(fn):
            return True

        # if it is not a parameter file, then it can't be of this type
        if not AHFCRMArbor._is_parameter_file(fn):
            return False

        mtree_fn = self._guess_crm_filename(self, fn)
        if mtree_fn is None or not os.path.exists(mtree_fn):
            return False

        return True



"""
Data structures for ytree frontend.




"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree Development Team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import h5py
import numpy as np
import json
import os

from yt.data_objects.static_output import \
    ParticleFile
from yt.frontends.ytdata.data_structures import \
    SavedDataset
from yt.geometry.particle_geometry_handler import \
    ParticleIndex

from ytree.utilities.io import parse_h5_attr
from ytree.yt_frontend.fields import YTreeFieldInfo

_ptype = "halos"
_unit_defaults = \
  {"mass":     {"field": "mass",       "units": "Msun"},
   "velocity": {"field": "velocity_x", "units": "km/s"},
   "time":     {"field": "time",       "units": "Gyr"}}

class YTreeHDF5File(ParticleFile):
    def __init__(self, ds, io, filename, file_id, frange):
        with h5py.File(filename, mode="r") as f:
            self.total_particles_file = \
              {_ptype: f['data'].attrs['num_elements']}
        super().__init__(ds, io, filename, file_id, frange)

    @property
    def _prefix(self):
        return self.filename[:-len(self.ds._suffix)]

    @property
    def _file_number(self):
        return int(self._prefix[-4:])

    @property
    def analysis_filename(self):
        return f"{self._prefix}-analysis{self.ds._suffix}"

    def _read_data(self, f, fname, mask):
        si = self.start
        ei = self.end
        return f[fname][si:ei][mask]

    def _get_file_root_index(self, f, mask):
        """
        Return index locations of the root node in this file.

        We use this to find the location of the root node in
        the arbor on the ytree side.
        """

        indices = np.arange(self.start, self.end)[mask]
        tree_start = f["index/tree_start_index"][()]
        return np.digitize(indices, tree_start) - 1

    def _get_file_number(self, mask):
        """
        Return the file number as a field-like array.

        We use this to find the index of the root node in the arbor.
        """

        return np.full(self.end-self.start, self._file_number)[mask]

    def _get_tree_index(self, f, mask):
        """
        Return the index of a halo in a tree.

        We use this to get the right node from a given root node.
        """

        indices = np.arange(self.start, self.end)[mask]
        tree_start = f["index/tree_start_index"][()]
        root_index = np.digitize(indices, tree_start) - 1
        return indices - tree_start[root_index]

    def _read_field_data(self, field, mask, f=None):
        if f is None:
            close = True
            f = h5py.File(self.filename, mode="r")
        else:
            close = False

        if self.ds._field_dict.get(field, {}).get("source") == "analysis":
            my_f = h5py.File(self.analysis_filename, mode="r")
            close = True
        else:
            my_f = f

        if field == "file_root_index":
            data = self._get_file_root_index(my_f, mask)
        elif field == "file_number":
            data = self._get_file_number(mask)
        elif field == "tree_index":
            data = self._get_tree_index(my_f, mask)
        else:
            data = self._read_data(my_f, os.path.join("data", field), mask)

        if close:
            my_f.close()

        return data

    def _get_particle_positions(self, ptype, f=None, transpose=True):
        if f is None:
            close = True
            f = h5py.File(self.filename, mode="r")
        else:
            close = False

        mask = slice(None)
        pn = "data/position_%s"
        with h5py.File(self.filename, mode="r") as f:
            units = parse_h5_attr(f[pn % "x"], "units")
            pos = np.vstack(
                [self._read_data(f, pn % ax, mask).astype("float64") for ax in "xyz"]).T

        if close:
            f.close()

        dle = self.ds.domain_left_edge.to(units).v
        dw = self.ds.domain_width.to(units).v

        np.subtract(pos, dle, out=pos)
        np.mod(pos, dw, out=pos)
        np.add(pos, dle, out=pos)
        if transpose:
            pos = pos.T
        pos = self.ds.arr(pos, units)

        return pos

class YTreeDataset(SavedDataset):
    _index_class = ParticleIndex
    _file_class = YTreeHDF5File
    _field_info_class = YTreeFieldInfo
    _suffix = ".h5"
    _con_attrs = ("hubble_constant", "omega_matter", "omega_lambda")
    _force_periodicity = True

    def __init__(self, filename, dataset_type="ytree_arbor",
                 index_order=None,
                 units_override=None, unit_system="cgs"):
        self._prefix = filename[:filename.rfind(self._suffix)]
        self.index_order = index_order
        super().__init__(filename, dataset_type,
                         units_override=units_override,
                         unit_system=unit_system)
        self._get_analysis_field_dict()

    def _get_analysis_field_dict(self):
        analysis_filename = f"{self._prefix}-analysis{self._suffix}"
        if not os.path.exists(analysis_filename):
            return

        with h5py.File(analysis_filename, mode="r") as f:
            afd = json.loads(f.attrs['field_info'])
        for fi in afd.values():
            fi["source"] = "analysis"
        self._field_dict.update(afd)

    def _set_derived_attrs(self):
        self.domain_center = 0.5 * (self.domain_right_edge +
                                    self.domain_left_edge)
        self.domain_width = self.domain_right_edge - self.domain_left_edge

    def _with_parameter_file_open(self, f):
        self.file_count = f.attrs['total_files']
        self.particle_count = f.attrs['total_nodes']
        self._field_dict = json.loads(f.attrs['field_info'])
        if "unit_system_name" not in self.parameters:
            self.parameters["unit_system_name"] = "mks"

    def _parse_parameter_file(self):
        self.current_redshift = None
        self.current_time = None
        self.cosmological_simulation = 1
        self.dimensionality = 3
        prefix = self.parameter_filename[:-len(self._suffix)]
        self.filename_template = f"{prefix}_%(num)04d{self._suffix}"
        self.particle_types = (_ptype)
        self.particle_types_raw = (_ptype)
        super()._parse_parameter_file()

        box_size = self.parameters["box_size"]
        self.domain_left_edge = box_size * np.zeros(self.dimensionality)
        self.domain_right_edge = box_size * np.ones(self.dimensionality)

        # Modify code units and attributes.
        self.unit_registry.modify("code_length", box_size.uq.in_base())
        self.parameters["length_unit"] = box_size.uq

        for myu, info in _unit_defaults.items():
            cu = self._field_dict.get(info["field"], info)["units"]
            cuval = self.quan(1, cu)
            self.parameters[f"{myu}_unit"] = cuval
            self.unit_registry.modify(f"code_{myu}", cuval.in_base())

    def __repr__(self):
        return self.basename[:self.basename.rfind(self._suffix)]

    @classmethod
    def _is_valid(self, *args, **kwargs):
        if not args[0].endswith(".h5"): return False
        with h5py.File(args[0], "r") as f:
            if "data_type" in f.attrs and "arbor_type" in f.attrs:
                return True
        return False



"""
tests for NodeContainer



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
from numpy.testing import assert_array_equal

import ytree

from ytree.utilities.testing import requires_file

TCL = "tiny_ctrees/locations.dat"

@requires_file(TCL)
def test_containers():
    a = ytree.load(TCL)

    first = a[0]
    last = a[-1]
    container = a.container([first, last])
    masses = a.arr([node["mass"] for node in [first, last]])
    assert_array_equal(container["mass"], masses)
    assert container.size == 2
    assert container.size == len(container)

    my_tree = a[0]
    leaves = my_tree.get_leaf_nodes()
    leaf_container = a.container(leaves)
    for i, leaf in enumerate(leaf_container):
        assert leaf["mass"] == leaf_container["mass"][i]

    a_slice = a.container(a[::8])
    assert int(np.ceil(a.size / 8)) == a_slice.size
    assert_array_equal(a_slice["mass"], a["mass"][::8])



from ytree.frontends.consistent_trees import \
    ConsistentTreesHlistArbor
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest

class ConsistentTreesHlistArborTest(TempDirTest, ArborTest):
    arbor_type = ConsistentTreesHlistArbor
    test_filename = "ctrees_hlists/hlists/hlist_0.12521.list"
    num_data_files = 10
    tree_skip = 100



"""
ytree frontend.




"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree Development Team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from ytree.yt_frontend.data_structures import YTreeDataset
from ytree.yt_frontend.io import IOHandlerYTreeHDF5



"""
ConsistentTreesArbor class and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import glob
import numpy as np
import operator
import os

from yt.funcs import \
    get_pbar

from ytree.data_structures.arbor import \
    SegmentedArbor

from ytree.frontends.consistent_trees.fields import \
    ConsistentTreesFieldInfo
from ytree.frontends.consistent_trees.io import \
    ConsistentTreesDataFile, \
    ConsistentTreesTreeFieldIO, \
    ConsistentTreesHlistDataFile
from ytree.frontends.consistent_trees.utilities import \
    parse_ctrees_header
from ytree.frontends.rockstar.arbor import \
    RockstarArbor

from ytree.utilities.exceptions import \
    ArborDataFileEmpty
from ytree.utilities.io import \
    f_text_block

class ConsistentTreesArbor(SegmentedArbor):
    """
    Arbors loaded from consistent-trees tree_*.dat files.
    """

    _field_info_class = ConsistentTreesFieldInfo
    _tree_field_io_class = ConsistentTreesTreeFieldIO
    _default_dtype = np.float32
    _node_io_attrs = ('_fi', '_si', '_ei')

    def _get_data_files(self):
        self.data_files = [ConsistentTreesDataFile(self.filename)]

    def _parse_parameter_file(self, filename=None,
                              ntrees_in_file=True):
        if filename is None:
            filename = self.filename

        fi = parse_ctrees_header(
            self, filename,
            ntrees_in_file=ntrees_in_file)
        self.field_list = list(fi.keys())
        self.field_info.update(fi)

    def _plant_trees(self):
        if self.is_planted or self._size == 0:
            return

        lkey = len("tree ")+1
        block_size = 4096

        data_file = self.data_files[0]

        data_file.open()
        data_file.fh.seek(0, 2)
        file_size = data_file.fh.tell()
        pbar = get_pbar("Loading tree roots", file_size)
        data_file.fh.seek(self._hoffset)

        offset = self._hoffset
        itree = 0
        nblocks = np.ceil(float(file_size-self._hoffset) /
                          block_size).astype(np.int64)
        for ib in range(nblocks):
            my_block = min(block_size, file_size - offset)
            if my_block <= 0: break
            buff = data_file.fh.read(my_block)
            lihash = -1
            for ih in range(buff.count("#")):
                ihash = buff.find("#", lihash+1)
                inl = buff.find("\n", ihash+1)
                if inl < 0:
                    buff += data_file.fh.readline()
                    inl = len(buff)
                uid = int(buff[ihash+lkey:inl])
                self._node_info['uid'][itree] = uid
                lihash = ihash
                self._node_info['_si'][itree] = offset + inl + 1
                self._node_info['_fi'][itree] = 0
                if itree > 0:
                    self._node_info['_ei'][itree-1] = offset + ihash - 1
                itree += 1
            offset = data_file.fh.tell()
            pbar.update(offset)
        self._node_info['_ei'][-1] = offset
        data_file.close()
        pbar.finish()

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        File should end in .dat and have a line in the header
        with the string, "Consistent Trees".
        """
        fn = args[0]
        if not fn.endswith(".dat"): return False
        with open(fn, "r") as f:
            valid = False
            while True:
                line = f.readline()
                if line is None or not line.startswith("#"):
                    break
                if "Consistent Trees" in line:
                    valid = True
                    break
            if not valid: return False
        return True

class ConsistentTreesGroupArbor(ConsistentTreesArbor):
    """
    Arbors loaded from consistent-trees locations.dat files.
    """

    def _get_data_files(self):
        pass

    def _parse_parameter_file(self):
        f = open(self.filename, 'r')
        f.readline()
        self._hoffset = f.tell()
        line = f.readline()
        if not line:
            raise ArborDataFileEmpty(self.filename)

        fn = os.path.join(self.directory, line.split()[3])
        super()._parse_parameter_file(filename=fn, ntrees_in_file=False)

    def _plant_trees(self):
        if self.is_planted:
            return

        f = open(self.filename, 'r')
        f.seek(self._hoffset)
        ldata = list(map(
            lambda x: [int(x[0]), int(x[1]), int(x[2]), x[3], len(x[0])],
            [line.split() for line, _ in f_text_block(f, pbar_string='Reading locations')]
            ))
        f.close()

        self._size = len(ldata)

        # It's faster to create and sort arrays and then sort ldata
        # for some reason.
        dfns = np.unique([datum[3] for datum in ldata])
        dfns.sort()
        fids = np.array([datum[1] for datum in ldata])
        fids.sort()
        ufids = np.unique(fids)
        ufids.sort()

        # Some data files may be empty and so unlisted.
        # Make sure file ids and names line up.
        data_files = [None]*(ufids.max()+1)
        for i,fid in enumerate(ufids):
            data_files[fid] = dfns[i]
        self.data_files = \
          [None if fn is None
           else ConsistentTreesDataFile(os.path.join(self.directory, fn))
           for fn in data_files]

        ldata.sort(key=operator.itemgetter(1, 2))
        pbar = get_pbar("Loading tree roots", self._size)

        # Set end offsets for each tree.
        # We don't get them from the location file.
        lkey = len("tree ")+3 # length of the separation line between trees
        same_file = np.diff(fids, append=fids[-1]+1) == 0

        for i, tdata in enumerate(ldata):
            self._node_info['uid'][i] = tdata[0]
            self._node_info['_fi'][i] = tdata[1]
            self._node_info['_si'][i] = tdata[2]
            # Get end index from next tree.
            if same_file[i]:
                self._node_info['_ei'][i] = ldata[i+1][2] - lkey - tdata[4]
            pbar.update(i+1)
        pbar.finish()

        # Get end index for last trees in files.
        for i in np.where(~same_file)[0]:
            data_file = self.data_files[fids[i]]
            data_file.open()
            data_file.fh.seek(0, 2)
            self._node_info['_ei'][i] = data_file.fh.tell()
            data_file.close()

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        File should end in .dat and have a line in the header
        with the string, "Consistent Trees".
        """
        fn = args[0]
        if not os.path.basename(fn) == 'locations.dat':
            return False
        with open(fn, "r") as f:
            valid = False
            while True:
                line = f.readline()
                if line is None or not line.startswith("#"):
                    break
                if "TreeRootID FileID Offset Filename" in line:
                    valid = True
            if not valid:
                return False
        return True

class ConsistentTreesHlistArbor(RockstarArbor):
    """
    Class for Arbors created from consistent-trees hlist_*.list files.

    This is a hybrid type with multiple catalog files like the rockstar
    frontend, but with headers structured like consistent-trees.
    """

    _has_uids = True
    _field_info_class = ConsistentTreesFieldInfo
    _data_file_class = ConsistentTreesHlistDataFile

    def _parse_parameter_file(self):
        ConsistentTreesArbor._parse_parameter_file(
            self, ntrees_in_file=False)

    def _get_data_files(self):
        """
        Get all out_*.list files and sort them in reverse order.
        """
        prefix = os.path.join(os.path.dirname(self.filename), "hlist_")
        suffix = ".list"
        my_files = glob.glob(f"{prefix}*{suffix}")

        # sort by catalog number
        my_files.sort(
            key=lambda x:
            self._get_file_index(x, prefix, suffix),
            reverse=True)
        self.data_files = \
          [self._data_file_class(f, self) for f in my_files]

    def _get_file_index(self, f, prefix, suffix):
        return float(f[f.find(prefix)+len(prefix):f.rfind(suffix)])

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        File should end in .list.
        """
        fn = args[0]
        if not os.path.basename(fn).startswith("hlist") or \
          not fn.endswith(".list"):
            return False
        return True



from ytree.data_structures.load import load as ytree_load
from ytree.frontends.treefrog import \
    TreeFrogArbor
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest

class TreeFrogArborTest(TempDirTest, ArborTest):
    arbor_type = TreeFrogArbor
    test_filename = "treefrog/VELOCIraptor.tree.t4.0-131.walkabletree.sage.forestID.foreststats.hdf5"
    groups = ("forest", "tree", "prog")
    num_data_files = 1
    tree_skip = 1000

    @property
    def arbor(self):
        arbor = super().arbor
        if "mass" not in arbor.field_info:
            arbor.add_alias_field(
                "mass", "Mass_200crit", units="Msun")
        return arbor

    def test_load_from_datafile(self):
        for df in self.arbor.data_files:
            new_arbor = ytree_load(df.filename)
            assert isinstance(new_arbor, self.arbor_type)



"""
ConsistentTreesArbor utility functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import re

from unyt.exceptions import \
    UnitParseError

def parse_ctrees_header(arbor, input_stream,
                        ntrees_in_file=True):
    """
    Parse consistent-trees header information.

    Parse ascii text from a file or list of strings to get:
    - cosmology parameters
    - box size
    - fields and units
    - optionally, number of trees and end of header file offset

    Return a dictionary of field information.
    """
    fields = []
    fi = {}
    fdb = {}
    rems = [f"{s[0]}{t}{s[1]}"
            for s in [("(", ")"), ("", "")]
            for t in ["physical, peculiar",
                      "comoving", "physical"]]

    if isinstance(input_stream, str):
        f = open(input_stream, mode='r')
        is_file = True
    else:
        is_file = False

    def next_line():
        if is_file:
            return f.readline()
        else:
            return input_stream.pop(0) \
              if input_stream else None

    # Read the first line as a list of all fields.
    # Do some footwork to remove awkard characters.
    rfl = next_line()[1:].strip().split()
    reg = re.compile(r"\(\d+\)$")
    for pf in rfl:
        match = reg.search(pf)
        if match is None:
            fields.append(pf)
        else:
            fields.append(pf[:match.start()])

    # Now grab a bunch of things from the header.
    while True:
        line = next_line()
        if line is None:
            if ntrees_in_file:
                raise IOError(
                    f"Encountered enexpected EOF reading {input_stream}.")
            else:
                break
        elif not line.startswith("#"):
            if ntrees_in_file:
                arbor._size = int(line.strip())
                arbor._hoffset = f.tell()
            break

        # cosmological parameters
        if "Omega_M" in line:
            pars = line[1:].split(";")
            for j, par in enumerate(["omega_matter",
                                     "omega_lambda",
                                     "hubble_constant"]):
                v = float(pars[j].split(" = ")[1])
                setattr(arbor, par, v)

        # box size
        elif "Full box size" in line:
            pars = line.split("=")[1].strip().split()
            box = pars

        # These are lines describing the various fields.
        # Pull them apart and look for units.
        elif ":" in line:
            tfields, desc = line[1:].strip().split(":", 1)

            # Units are enclosed in parentheses.
            # Pull out what's enclosed and remove things like
            # "comoving" and "physical".
            if "(" in line and ")" in line:
                punits = desc[desc.find("(")+1:desc.rfind(")")]
                for rem in rems:
                    while rem in punits:
                        pre, mid, pos = punits.partition(rem)
                        punits = pre + pos
                try:
                    arbor.quan(1, punits)
                except UnitParseError:
                    punits = ""
            else:
                punits = ""

            # Multiple fields together on the same line.
            for sep in ["/", ","]:
                if sep in tfields:
                    tfields = tfields.split(sep)
                    break
            if not isinstance(tfields, list):
                tfields = [tfields]

            # Assign units and description.
            for tfield in tfields:
                fdb[tfield.lower()] = {"description": desc.strip(),
                                       "units": punits}

    if is_file:
        f.close()

    # Fill the field info with the units found above.
    for i, field in enumerate(fields):
        if "(" in field and ")" in field:
            cfield = field[:field.find("(")]
        else:
            cfield = field
        fi[field] = fdb.get(cfield.lower(),
                            {"description": "",
                             "units": ""})
        fi[field]["column"] = i

    arbor.box_size = arbor.quan(float(box[0]), box[1])
    return fi



from ytree.frontends.consistent_trees_hdf5 import \
    ConsistentTreesHDF5Arbor
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest

class ConsistentTreesHDF5ArborTest4(TempDirTest, ArborTest):
    arbor_type = ConsistentTreesHDF5Arbor
    test_filename = ["consistent_trees_hdf5/soa/forest_0.h5",
                     "consistent_trees_hdf5/soa/forest_0.h5"]
    num_data_files = 2
    tree_skip = 20000



"""
AnalysisPipeline class and member functions



"""

import os

from yt.utilities.parallel_tools.parallel_analysis_interface import parallel_root_only

from ytree.analysis.analysis_operators import AnalysisOperation
from ytree.utilities.io import ensure_dir

class AnalysisPipeline:
    """
    Initialize an AnalysisPipeline.

    An AnalysisPipeline allows one to create a workflow of analysis to be
    performed on a node/halo in a tree. This is done by creating functions
    that minimally accept a node as the first argument and providing these
    to the AnalysisPipeline in the order they are meant to be run. This
    makes it straightforward to organize an analysis workflow into a series
    of distinct, reusable functions.

    Parameters
    ----------
    output_dir : optional, str
        Path to a directory into which any files will be saved. The
        directory will be created if it does not already exist.

    Examples
    --------

    >>> import ytree
    >>>
    >>> def my_analysis(node):
    ...     node["test_field"] = 2 * node["mass"]
    >>>
    >>> def minimum_mass(node, value):
    ...     return node["mass"] > value
    >>>
    >>> def my_recipe(pipeline):
    ...     pipeline.add_operation(my_analysis)
    >>>
    >>> def do_cleanup(node):
    ...     print (f"End of analysis for {node}.")
    >>>
    >>> a = ytree.load("arbor/arbor.h5")
    >>>
    >>> ap = AnalysisPipeline()
    >>> # don't analyze halos below 3e11 Msun
    >>> ap.add_operation(minimum_mass, 3e11)
    >>> ap.add_recipe(my_recipe)
    >>> ap.add_recipe(do_cleanup, always_do=True)
    >>>
    >>> trees = list(a[:])
    >>> for tree in trees:
    ...     for node in tree["forest"]:
    ...         ap.process_target(node)
    >>>
    >>> a.save_arbor(trees=trees)
    """

    def __init__(self, output_dir=None):
        self.actions = []
        if output_dir is None:
            output_dir = "."
        self.output_dir = ensure_dir(output_dir)
        self._preprocess_functions = []
        self._preprocessed = False
        self._handoff_store = {}

    def add_operation(self, function, *args, always_do=False,
                      preprocess_function=None, **kwargs):
        """
        Add an operation to the AnalysisPipeline.

        An operation is a function that minimally takes in a target object
        and performs some actions on or with it. This function may alter the
        object's state, attach attributes, write out data, etc. Operations
        are used to create a pipeline of actions performed in sequence on a list
        of objects, such as all halos in a merger tree. The function can,
        optionally, return True or False to act as a filter, determining if the
        rest of the pipeline should be carried out (if True) or if the pipeline
        should stop and move on to the next object (if False).

        Parameters
        ----------
        function : callable
            The function to be called for each node/halo.
        *args : positional arguments
            Any additional positional arguments to be provided to the funciton.
        always_do : optional, bool
            If True, always perform this operation even if a prior filter has
            returned False. This can be used to add house cleaning operations
            that should always be run.
            Default: False
        preprocess_function : optional, callable
            A function to be run once at the start of analysis. This can be
            used to perform any necessary startup tasks prior to running the
            pipeline. The function provided must accept no arguments.
            Default: None
        **kwargs : keyword arguments
            Any keyword arguments to be provided to the function.
        """

        if not callable(function):
            raise ValueError("function argument must be a callable function.")

        operation = AnalysisOperation(function, *args, always_do=always_do, **kwargs)
        self.actions.append(operation)
        if preprocess_function is not None:
            self._preprocess_functions.append(preprocess_function)

    def add_recipe(self, function, *args, **kwargs):
        """
        Add a recipe to the AnalysisPipeline.

        An recipe is a function that accepts an AnalysisPipeline and adds a
        series of operations with calls to add_operation. This is a way of
        creating a shortcut for a series of operations.

        Parameters
        ----------
        function : callable
            A function accepting an AnalysisPipeline object.
        *args : positional arguments
            Any additional positional arguments to be provided to the funciton.
        **kwargs : keyword arguments
            Any keyword arguments to be provided to the function.

        Examples
        --------
        >>> def print_field_value(node, field):
        ...     print (f"Node {node} has {field} of {node[field]}.")
        >>>
        >>> def print_many_things(pipeline, fields):
        ...    for field in fields:
        ...        pipeline.add_operation(print_field_value, field)
        >>>
        >>> ap = ytree.AnalysisPipeline()
        >>> ap.add_recipe(print_many_things, ["mass", "virial_radius"])
        """

        if not callable(function):
            raise ValueError("function argument must be a callable function.")

        recipe = AnalysisOperation(function, *args, **kwargs)
        recipe(self)

    @parallel_root_only
    def _preprocess(self):
        """
        Create output directories and do any other preliminary steps.

        Run any preprocess functions that were added.
        """

        if self._preprocessed:
            return

        for pre_func in self._preprocess_functions:
            pre_func()

        for action in self.actions:
            my_output_dir = action.kwargs.get("output_dir")
            if my_output_dir is not None:
                new_output_dir = ensure_dir(
                    os.path.join(self.output_dir, my_output_dir))
                action.kwargs["output_dir"] = new_output_dir

        self._preprocessed = True

    def process_target(self, target, handoff_attrs=None):
        """
        Process a node through the AnalysisPipeline.

        All operations added to the AnalysisPipeline will be run on the
        provided target.

        Parameters
        ----------
        target : :class:`~ytree.data_structures.tree_node.TreeNode`
            The node on which to run the analysis pipeline.
        handoff_attrs : optional, list of strings
            A list of attributes to be handed down from one target to
            the next. If given, these attributes will be taken from the
            target object after the pipeline is run and stored internally.
            They will then be attached to the next target run through the
            pipeline. This can be used to pass down attributes from
            one target to the next, for example, to accumulate results
            or hang onto objects that are expensive to create.
            Default: None.
        """

        self._preprocess()
        if handoff_attrs is None:
            handoff_attrs = []

        for attr in handoff_attrs:
            if attr in self._handoff_store:
                val = self._handoff_store.pop(attr)
                setattr(target, attr, val)

        target_filter = True
        for action in self.actions:
            if target_filter or action.always_do:
                rval = action(target)
                if rval is not None:
                    target_filter &= bool(rval)

        for attr in handoff_attrs:
            if hasattr(target, attr):
                val = getattr(target, attr)
                self._handoff_store[attr] = val
                delattr(target, attr)

        return target_filter



import h5py
import numpy as np
from numpy.testing import assert_array_equal

from ytree.frontends.moria import \
    MoriaArbor
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest

class MoriaArborTest(TempDirTest, ArborTest):
    arbor_type = MoriaArbor
    test_filename = "moria/moria_tree_testsim050.hdf5"
    groups = ("forest", "tree", "prog")
    num_data_files = 1
    tree_skip = 100

    @property
    def arbor(self):
        arbor = super().arbor
        if "mass" not in arbor.field_info:
            arbor.add_alias_field(
                "mass", "Mpeak", units="Msun")
        return arbor

    def test_missing_descuids(self):
        a = self.arbor
        a._plant_trees()

        fh = h5py.File(self.arbor.parameter_filename, mode="r")
        dids = fh['descendant_id'][()]
        ids = fh['id'][()]
        dindex = fh['descendant_index'][()]
        missing = np.setdiff1d(dids, ids)

        for my_desc in missing:
            if my_desc == -1:
                continue

            index = np.where(dids == my_desc)
            my_id = ids[index][0]
            my_dindex = (index[0]+1, dindex[index][0])

            ai = np.digitize(dindex[index][0], a._node_info["_si"]) - 1
            t = a[ai]
            inode = t["forest", "uid"] == my_id
            assert_array_equal(t["forest", "desc_uid"][inode], ids[my_dindex])



"""
tests for derived fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
from numpy.testing import \
    assert_array_equal, \
    assert_equal

from ytree.utilities.testing import \
    requires_file, \
    TempDirTest

import ytree

CT = "consistent_trees/tree_0_0_0.dat"

def _potential(field, data):
    return data['mass'] / data['virial_radius']

def _uid_mod7(field, data):
    return data['uid'] % 7

class DerivedFieldTest(TempDirTest):
    @requires_file(CT)
    def test_derived_fields(self):
        a = ytree.load(CT)

        a.add_derived_field('potential', _potential,
                             units='Msun/kpc')

        m = a['mass'].copy()
        r = a['virial_radius'].copy()
        assert_array_equal(a['potential'], m/r)
        assert_equal(str(a['potential'].units), 'Msun/kpc')
        assert_equal(a['potential'].dtype, a._default_dtype)

        my_tree = a[0]
        m = my_tree['mass'].copy()
        r = my_tree['virial_radius'].copy()
        assert_array_equal(my_tree['potential'], m/r)
        assert_equal(str(my_tree['potential'].units), 'Msun/kpc')
        assert_equal(my_tree['potential'].dtype, a._default_dtype)

    @requires_file(CT)
    def test_derived_fields_int(self):
        a = ytree.load(CT)

        my_dtype = np.int64
        a.add_derived_field('uid_mod7', _uid_mod7,
                            units='', dtype=my_dtype)

        um7 = a['uid'] % 7
        assert_array_equal(a['uid_mod7'], um7)
        assert_equal(a['uid_mod7'].dtype, my_dtype)

        my_tree = a[0]
        um7 = my_tree['tree', 'uid'] % 7
        assert_array_equal(my_tree['tree', 'uid_mod7'], um7)
        assert_equal(my_tree['tree', 'uid_mod7'].dtype, my_dtype)



"""
miscellaneous utilities



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

def fround(val, decimals=0):
    """
    Round 0.5 up to 1 and so on.
    """
    fac = 10**decimals
    return np.floor(val * fac + 0.5) / fac



"""
ytree-specific fields




"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree Development Team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.fields.field_info_container import \
    FieldInfoContainer

p_units = 'unitary'
v_units = 'km/s'

def _redshift(field, data):
    ptype, _ = field.name
    return 1. / data[ptype, 'scale_factor'] - 1

def _time(field, data):
    ptype, _ = field.name
    return data.ds.cosmology.t_from_z(data[ptype, "redshift"])

class YTreeFieldInfo(FieldInfoContainer):
    known_other_fields = (
    )

    known_particle_fields = (
        ("position_x", (p_units, ('halos', 'particle_position_x'), None)),
        ("position_y", (p_units, ('halos', 'particle_position_y'), None)),
        ("position_z", (p_units, ('halos', 'particle_position_z'), None)),
        ("velocity_x", (v_units, ('halos', 'particle_velocity_x'), None)),
        ("velocity_y", (v_units, ('halos', 'particle_velocity_y'), None)),
        ("velocity_z", (v_units, ('halos', 'particle_velocity_z'), None)),
        ("mass", ('Msun', ('halos', 'particle_mass'), None)),
    )

    def setup_particle_fields(self, ptype):
        if (ptype, "redshift") not in self.ds.field_list:
            self.add_field((ptype, 'redshift'), function=_redshift,
                           sampling_type='particle', units='')

        if (ptype, "time") not in self.ds.field_list:
            self.add_field((ptype, 'time'), function=_time,
                           sampling_type='particle', units='Gyr')

        super().setup_particle_fields(ptype)



"""
ConsistentTreesArbor fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from ytree.data_structures.fields import \
    FieldInfoContainer

m_unit = "Msun"
p_unit = "unitary"
r_unit = "kpc"
v_unit = "km/s"

id_type = np.int64

class ConsistentTreesFieldInfo(FieldInfoContainer):
    alias_fields = (
        ("uid", "id", None),
        ("desc_uid", "desc_id", None),
        ("halo_id", "Orig_halo_ID", None),
        ("scale_factor", "scale", None),
        ("mass", "Mvir", m_unit),
        ("virial_mass", "Mvir", m_unit),
        ("virial_radius", "Rvir", r_unit),
        ("scale_radius", "rs", r_unit),
        ("velocity_dispersion", "vrms", v_unit),
        ("position_x", "x", p_unit),
        ("position_y", "y", p_unit),
        ("position_z", "z", p_unit),
        ("velocity_x", "vx", v_unit),
        ("velocity_y", "vy", v_unit),
        ("velocity_z", "vz", v_unit),
        ("angular_momentum_x", "Jx", None),
        ("angular_momentum_y", "Jy", None),
        ("angular_momentum_z", "Jz", None),
        ("spin_parameter", "Spin", None),
    )

    data_types = (
        ("id", id_type),
        ("desc_id", id_type),
        ("num_prog", np.int32),
        ("pid", id_type),
        ("upid", id_type),
        ("desc_pid", id_type),
        ("Breadth_first_ID", id_type),
        ("Depth_first_ID", id_type),
        ("Tree_root_ID", id_type),
        ("Orig_halo_ID", id_type),
        ("Snap_idx", id_type),
        ("Next_coprogenitor_depthfirst_ID", id_type),
        ("Last_progenitor_depthfirst_ID", id_type),
        ("Last_mainleaf_depthfirst_ID", id_type),
        ("Tidal_ID", id_type),
    )



"""
AHFArbor miscellany



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

def parse_AHF_file(filename, pars, sep=None):
    """
    Parse an AHF log or parameter file.
    """

    npars = len(pars.keys())
    vals = {}

    f = open(filename, "r")
    for line in f.readlines():
        line = line.strip()
        if line is None:
            break
        if len(vals.keys()) == npars:
            break
        for par in pars:
            key = pars[par]
            if key in vals:
                continue
            if line.startswith(f"{par} "):
                val = float(line.split(sep)[1])
                vals[key] = val
    f.close()

    if len(vals) < npars:
        mpars = ", ".join(set(pars.values()).difference(vals))
        raise RuntimeError(f"{filename} missing these parameters: {mpars}.")

    return vals



"""
io utilities



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import errno
import numpy as np
import os
from unyt import \
    unyt_array, \
    unyt_quantity

from yt.funcs import \
    get_pbar
from ytree.utilities.logger import \
    fake_pbar

def dirname(path, level=1):
    """
    Multi-level version of os.path.dirname.
    """
    if not isinstance(level, int) or level < 1:
        raise ValueError(
            f"level must be a positive integer: {level}.")
    for i in range(level):
        path = os.path.dirname(path)
    return path

def ensure_dir(path):
    r"""Parallel safe directory maker."""
    if os.path.exists(path):
        return path

    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise
    return path

def parse_h5_attr(f, attr):
    """A Python3-safe function for getting hdf5 attributes.

    If an attribute is supposed to be a string, this will return it as such.

    This was taken from yt.
    """
    val = f.attrs.get(attr, None)
    if isinstance(val, bytes):
        return val.decode('utf8')
    else:
        return val

def _hdf5_yt_attr(fh, attr, unit_registry=None):
    """
    Read an hdf5 attribute.  If there exists another attribute
    named <attr>_units, use that to assign units and return
    as either a unyt_array or unyt_quantity.
    """
    val = fh.attrs[attr]
    units = ""
    ufield = f"{attr}_units"
    if ufield in fh.attrs:
        units = fh.attrs[ufield]
        if isinstance(units, bytes):
            units = units.decode("utf")
    if units == "dimensionless":
        units = ""
    if units != "":
        if isinstance(val, np.ndarray):
            val = unyt_array(val, units, registry=unit_registry)
        else:
            val = unyt_quantity(val, units, registry=unit_registry)
    return val

def _hdf5_yt_array_lite(fh, field):
    """
    Read an hdf5 dataset.  If that dataset has a "units" attribute,
    return that as well, but do not cast as a unyt_array.
    """
    units = ""
    if "units" in fh[field].attrs:
        units = fh[field].attrs["units"]
    if units == "dimensionless": units = ""
    return (fh[field][()], units)

def f_text_block(f, block_size=4096, file_size=None, sep="\n",
                 pbar_string=None):
    """
    Read lines from a file faster than f.readlines().
    """
    start = f.tell()
    if file_size is None:
        f.seek(0, 2)
        file_size = f.tell() - start
        f.seek(start)

    nblocks = np.ceil(float(file_size) /
                      block_size).astype(np.int64)
    read_size = file_size + start
    lbuff = ""
    if pbar_string is None:
        pbar = fake_pbar()
    else:
        pbar = get_pbar(pbar_string, file_size)
    for ib in range(nblocks):
        offset = f.tell()
        my_block = min(block_size, read_size-offset)
        if my_block <= 0: break
        buff = f.read(my_block)
        linl = -1
        for ih in range(buff.count(sep)):
            inl = buff.find(sep, linl+1)
            if inl < 0:
                lbuff += buff[linl+1:]
                continue
            else:
                line = lbuff + buff[linl+1:inl]
                loc = offset - len(lbuff) + linl + 1
                lbuff = ""
                linl = inl
                pbar.update(loc+len(line)-start+1)
                yield line, loc
        lbuff += buff[linl+1:]
    if lbuff:
        loc = f.tell() - len(lbuff)
        pbar.update(loc+len(lbuff)-start+1)
        yield lbuff, loc
    pbar.finish()



"""
Gadget4Arbor io classes and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import h5py
import numpy as np
import re

from ytree.data_structures.io import \
    DataFile, \
    TreeFieldIO

class Gadget4DataFile(DataFile):
    _io_attrs = ("ntrees", "nnodes", "tree_sizes", "offsets")
    def _load_properties(self):
        self.open()
        self.ntrees = int(self.fh["Header"].attrs["Ntrees_ThisFile"])
        self.nnodes = int(self.fh["Header"].attrs["Nhalos_ThisFile"])
        if self.ntrees > 0:
            self.tree_sizes = self.fh["TreeTable/Length"][()]
            self.offsets = self.fh["TreeTable/StartOffset"][()]
        else:
            self.tree_sizes = None
            self.offsets = None
        self.close()

    def __getattr__(self, attr):
        if attr in self._io_attrs:
            self._load_properties()
            return getattr(self, attr)
        raise AttributeError(attr)

    def open(self):
        self.fh = h5py.File(self.filename, mode="r")

    def close(self):
        self.fh.close()
        self.fh = None

class Gadget4TreeFieldIO(TreeFieldIO):
    def _read_fields(self, root_node, fields, dtypes=None,
                     root_only=False):
        """
        Read fields from disk for a single tree.
        """

        fi = self.arbor.field_info
        afields = [field for field in fields
                   if fi[field].get("source") == "arbor"]
        rfields = list(set(fields).difference(afields))

        for afield in afields:
            rfields.extend(
                [dfield for dfield in fi[afield].get("dependencies", [])
                 if dfield not in rfields])

        if root_only:
            fei = root_node._fi
        else:
            fei = root_node._fei

        freg = re.compile(r"(^.+)_(\d+$)")
        field_data = {field: [] for field in rfields}
        for dfi in range(root_node._fi, fei+1):
            data_file = self.arbor.data_files[dfi]
            close = False
            if data_file.fh is None:
                close = True
                data_file.open()
            fh = data_file.fh
            g = fh["TreeHalos"]

            si = root_node._si
            if root_only:
                my_slice = slice(si, si+1)
            else:
                if dfi == root_node._fi:
                    my_start = si
                else:
                    my_start = None

                if dfi == fei:
                    my_end = root_node._ei
                else:
                    my_end = None

                my_slice = slice(my_start, my_end)

            field_cache = {}
            for field in rfields:
                fs = freg.search(field)
                if fs and fs.groups()[0] in g:
                    fieldname, ifield = fs.groups()
                    ifield = int(ifield)
                    if fieldname not in field_cache:
                        field_cache[fieldname] = g[fieldname][my_slice]
                    field_data[field].append(field_cache[fieldname][:, ifield])
                else:
                    field_data[field].append(g[field][my_slice])

            if close:
                data_file.close()

        for field in field_data:
            field_data[field] = np.concatenate(field_data[field])

        if afields:
            field_data.update(self._get_arbor_fields(
                root_node, field_data, fields, afields, root_only))

        self._apply_units(rfields, field_data)

        return field_data

    def _get_arbor_fields(self, root_node, field_data,
                          fields, afields, root_only):
        """
        Generate special fields from the arbor/treenode.
        """

        adata = {}

        if "uid" in afields:
            if root_only:
                adata["uid"] = np.array([root_node.uid])
            else:
                adata["uid"] = root_node.uid + \
                  np.arange(root_node._tree_size)

        if "desc_uid" in afields:
            if "TreeDescendant" in fields:
                desc_uids = field_data["TreeDescendant"].copy()
            else:
                desc_uids = field_data.pop("TreeDescendant")
            desc_uids[desc_uids != -1] += root_node.uid
            adata["desc_uid"] = desc_uids

        return adata



"""
Gadget4Arbor fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import re

from ytree.data_structures.fields import \
    FieldInfoContainer

m_unit = "1e10 * Msun/h"
p_unit = "Mpc/h"
v_unit = "km/s"
j_unit = f"{p_unit} * {v_unit}"

def _redshift(field, data):
    isn = data["SnapNum"].astype(int)
    return data.arbor.arr(data.arbor._redshifts[isn], "")

class Gadget4FieldInfo(FieldInfoContainer):
    alias_fields = (
        ("group_mass", "Group_M_Crit200", m_unit),
        ("mass", "SubhaloMass", m_unit),
        ("position_x", "SubhaloPos_0", p_unit),
        ("position_y", "SubhaloPos_1", p_unit),
        ("position_z", "SubhaloPos_2", p_unit),
        ("velocity_x", "SubhaloVel_0", v_unit),
        ("velocity_y", "SubhaloVel_1", v_unit),
        ("velocity_z", "SubhaloVel_2", v_unit),
        ("angular_momentum_x", "SubhaloSpin_0", j_unit),
        ("angular_momentum_y", "SubhaloSpin_1", j_unit),
        ("angular_momentum_z", "SubhaloSpin_2", j_unit),
        ("velocity_dispersion", "SubhaloVelDisp", v_unit),
    )

    known_fields = (
        ("Group_M_Crit200", m_unit),
        ("SubhaloHalfmassRad", p_unit),
        ("SubhaloMass", m_unit),
        ("SubhaloPos", p_unit),
        ("SubhaloSpin", j_unit),
        ("SubhaloVel", v_unit),
        ("SubhaloVelDisp", v_unit),
        ("SubhaloVmax", v_unit),
        ("SubhaloVmaxRad", p_unit),
    )

    def setup_known_fields(self):
        """
        Add units for all <fieldname>_<number> fields as well.
        """

        kfields = dict(self.known_fields)
        freg = re.compile(r"(^.+)_\d+$")
        for field in self:
            if self[field].get("units") is not None:
                continue

            if field in kfields:
                self[field]["units"] = kfields[field]
                continue

            fs = freg.search(field)
            if fs and fs.groups()[0] in kfields:
                self[field]["units"] = kfields[fs.groups()[0]]

    def setup_derived_fields(self):
        self.arbor.add_derived_field(
            "redshift", _redshift, units="", force_add=False)
        # this will add it to field list for saving
        self.arbor.field_list.append("redshift")

        super().setup_derived_fields()



"""
TreeFrogArbor fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from ytree.data_structures.fields import \
    FieldInfoContainer

m_unit = "Msun"
p_unit = "unitary"
r_unit = "kpc"
v_unit = "km/s"

id_type = np.int64

class TreeFrogFieldInfo(FieldInfoContainer):
    alias_fields = (
        ("uid", "ID", ""),
        ("desc_uid", "Descendant", ""),
        ("velocity_dispersion", "sigV", v_unit),
        ("position_x", "Xc", p_unit),
        ("position_y", "Yc", p_unit),
        ("position_z", "Zc", p_unit),
        ("velocity_x", "VXc", v_unit),
        ("velocity_y", "VYc", v_unit),
        ("velocity_z", "VZc", v_unit),
        ("angular_momentum_x", "Lx", None),
        ("angular_momentum_y", "Ly", None),
        ("angular_momentum_z", "Lz", None),
        ("spin_parameter", "lambda_B", None),
    )

    known_fields = (
        ("Lx", "angular_momentum_unit"),
        ("Ly", "angular_momentum_unit"),
        ("Lz", "angular_momentum_unit"),
        ("Mass_200crit", "mass_unit"),
        ("Mass_200mean", "mass_unit"),
        ("Mass_FOF", "mass_unit"),
        ("Mass_tot", "mass_unit"),
        ("RVmax_Lx", "length_unit"),
        ("RVmax_Ly", "length_unit"),
        ("RVmax_Lz", "length_unit"),
        ("RVmax_sigV", "length_unit"),
        ("R_200crit", "length_unit"),
        ("R_200mean", "length_unit"),
        ("R_HalfMass", "length_unit"),
        ("R_size", "length_unit"),
        ("Rmax", "length_unit"),
        ("VXc", "velocity_unit"),
        ("VYc", "velocity_unit"),
        ("VZc", "velocity_unit"),
        ("Vmax", "velocity_unit"),
        ("Xc", "length_unit"),
        ("Xcminpot", "length_unit"),
        ("Yc", "length_unit"),
        ("Ycminpot", "length_unit"),
        ("Zc", "length_unit"),
        ("Zcminpot", "length_unit"),
        ("sigV", "velocity_unit"),
    )

    data_types = (
        ("ID", id_type),
        ("Descendant", id_type),
    )

    def __init__(self, arbor):
        super().__init__(arbor)
        self._setup_field_units()

    def _setup_field_units(self):
        a = self.arbor
        units = {}
        units["length_unit"] = a.quan(a.units["Length_unit_to_kpc"], "kpc")
        units["mass_unit"] = a.quan(a.units["Mass_unit_to_solarmass"], "Msun")
        units["velocity_unit"] = a.quan(a.units["Velocity_unit_to_kms"], "km/s")
        units["angular_momentum_unit"] = units["mass_unit"] * \
          units["length_unit"] * units["velocity_unit"]

        for unit, val in units.items():
            units[unit] = f"{str(val.d)}*{str(val.units)}"

        known_fields = []
        for field, ustr in self.known_fields:
            known_fields.append((field, units.get(ustr, "")))
        self.known_fields = tuple(known_fields)



"""
tests for analysis fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
from numpy.testing import \
    assert_array_equal, \
    assert_equal, \
    assert_raises

from ytree.utilities.exceptions import \
    ArborFieldAlreadyExists, \
    ArborUnsettableField
from ytree.utilities.testing import \
    requires_file, \
    TempDirTest

import ytree

CT = "consistent_trees/tree_0_0_0.dat"

class AnalysisFieldTest(TempDirTest):
    @requires_file(CT)
    def test_analysis_fields(self):
        a = ytree.load(CT)

        a.add_analysis_field('bears', 'Msun/yr')
        assert_equal(a._default_dtype, a['bears'].dtype)

        fake_bears = np.zeros(a.size)
        assert_array_equal(fake_bears, a['bears'])

        all_trees = list(a[:])
        my_tree = all_trees[0]
        my_tree['bears'] = 9
        fake_bears[0] = 9
        assert_array_equal(fake_bears, a['bears'])

        my_tree = all_trees[1]
        assert_equal(a._default_dtype, my_tree['tree', 'bears'].dtype)
        fake_tree_bears = np.zeros(my_tree.tree_size)
        assert_array_equal(
            fake_tree_bears, my_tree['tree', 'bears'])
        fake_tree_bears[72] = 99
        my_halos = list(my_tree['tree'])
        my_halos[72]['bears'] = 99
        assert_array_equal(fake_tree_bears, my_tree['tree', 'bears'])

        fn = a.save_arbor(trees=all_trees)
        a2 = ytree.load(fn)

        assert_array_equal(fake_bears, a2['bears'])
        assert_array_equal(fake_tree_bears, a2[1]['tree', 'bears'])
        assert_equal(a2['bears'].dtype, a._default_dtype)
        assert_equal(a2[1]['tree', 'bears'].dtype, a._default_dtype)

    @requires_file(CT)
    def test_analysis_fields_int(self):
        a = ytree.load(CT)

        my_dtype = np.int64
        a.add_analysis_field('bears', 'Msun/yr', dtype=my_dtype)
        assert_equal(my_dtype, a['bears'].dtype)

        fake_bears = np.zeros(a.size)
        assert_array_equal(fake_bears, a['bears'])

        all_trees = list(a[:])
        my_tree = all_trees[0]
        my_tree['bears'] = 9
        fake_bears[0] = 9
        assert_array_equal(fake_bears, a['bears'])

        my_tree = all_trees[1]
        assert_equal(my_dtype, my_tree['tree', 'bears'].dtype)
        fake_tree_bears = np.zeros(my_tree.tree_size)
        assert_array_equal(
            fake_tree_bears, my_tree['tree', 'bears'])
        fake_tree_bears[72] = 99
        my_halos = list(my_tree['tree'])
        my_halos[72]['bears'] = 99
        assert_array_equal(fake_tree_bears, my_tree['tree', 'bears'])

        fn = a.save_arbor(trees=all_trees)
        a2 = ytree.load(fn)

        assert_array_equal(fake_bears, a2['bears'])
        assert_array_equal(fake_tree_bears, a2[1]['tree', 'bears'])
        assert_equal(a2['bears'].dtype, my_dtype)
        assert_equal(a2[1]['tree', 'bears'].dtype, my_dtype)

    @requires_file(CT)
    def test_analysis_field_already_exists(self):
        a = ytree.load(CT)

        with assert_raises(ArborFieldAlreadyExists):
            a.add_analysis_field('mass', units='g')

    @requires_file(CT)
    def test_analysis_unsettable_fields(self):
        a = ytree.load(CT)

        with assert_raises(ArborUnsettableField):
            my_tree = a[0]
            my_tree['mass'] = 50

    @requires_file(CT)
    def test_add_vector_field(self):
        a = ytree.load(CT)

        data = []
        for i, ax in enumerate('xyz'):
            a.add_analysis_field(f'thing_{ax}', '', default=i, dtype=np.float64)
            data.append(i*np.ones(a.size, dtype=np.float64))
        data = np.rollaxis(np.vstack(data), 1)

        fn = a.save_arbor()
        a2 = ytree.load(fn)
        a2.add_vector_field('thing')

        assert_array_equal(a2['thing'], data)

        data_mag = np.sqrt((data**2).sum(axis=1))
        assert_array_equal(a2['thing_magnitude'], data_mag)

    @requires_file(CT)
    def test_alter_vector_field(self):
        a = ytree.load(CT)

        for i, ax in enumerate('xyz'):
            a.add_analysis_field(f'thing_{ax}', '', default=i, dtype=np.float64)

        fn = a.save_arbor()
        a2 = ytree.load(fn)
        a2.add_vector_field('thing')

        t = a2[0]
        for i, ax in enumerate('xyz'):
            t[f'thing_{ax}'] += 1
            assert_array_equal(t['forest', f'thing_{ax}'], t['forest', 'thing'][:, i])
            assert_array_equal(a2[f'thing_{ax}'], a2['thing'][:, i])

    @requires_file(CT)
    def test_reload_arbor(self):
        a = ytree.load(CT)

        fn = a.save_arbor()
        a = ytree.load(fn)
        a.add_analysis_field('bears', 'Msun/yr')

        trees = list(a[:])
        a.save_arbor(trees=trees, save_in_place=True)

        a = a.reload_arbor()
        assert "bears" in a.field_list



"""
RockstarArbor io classes and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from ytree.data_structures.io import \
    CatalogDataFile
from ytree.utilities.io import \
    f_text_block

class RockstarDataFile(CatalogDataFile):
    def __init__(self, filename, arbor):
        self.offsets = None
        super().__init__(filename, arbor)

    def open(self):
        self.fh = open(self.filename, "r")

    def _parse_header(self):
        self.open()
        f = self.fh
        f.seek(0, 2)
        self.file_size = f.tell()
        f.seek(0)
        while True:
            line = f.readline()
            if line is None:
                self._hoffset = f.tell()
                break
            elif not line.startswith("#"):
                self._hoffset = f.tell() - len(line)
                break
            elif line.startswith("#a = "):
                self.scale_factor = float(line.split(" = ")[1])
        self.close()

    def _read_data_default(self, rfields, dtypes):
        if not rfields:
            return {}

        fi = self.arbor.field_info
        field_data = \
          self._create_field_arrays(rfields, dtypes)
        offsets = []

        self.open()
        f = self.fh
        f.seek(self._hoffset)
        file_size = self.file_size - self._hoffset
        for line, offset in f_text_block(f, file_size=file_size):
            offsets.append(offset)
            sline = line.split()
            for field in rfields:
                field_data[field].append(sline[fi[field]["column"]])
        self.close()

        for field in rfields:
            field_data[field] = \
              np.array(field_data[field], dtype=dtypes[field])

        if self.offsets is None:
            self.offsets = np.array(offsets)

        return field_data

    def _read_data_select(self, rfields, tree_nodes, dtypes):
        if not rfields:
            return {}

        fi = self.arbor.field_info
        nt = len(tree_nodes)
        field_data = \
          self._create_field_arrays(rfields, dtypes, size=nt)

        self.open()
        f = self.fh
        for i in range(nt):
            f.seek(self.offsets[tree_nodes[i]._fi])
            line = f.readline()
            sline = line.split()
            for field in rfields:
                dtype = dtypes[field]
                field_data[field][i] = dtype(sline[fi[field]["column"]])
        self.close()

        return field_data



"""
YTreeArbor utilities



"""

_ops = {"above": ">=", "below": "<=", "equal": "=="}

def get_conditional(op, criterion):
    my_op = _ops[op]

    if len(criterion) == 2:
        field, value = criterion
        units = None
    elif len(criterion) == 3:
        field, value, units = criterion
    else:
        raise ValueError(
            "Criterion must have either two or three values: "
            f"{criterion}.")

    if units is None:
        ustr = ""
    else:
        ustr = f".in_units(\"{units}\")"

    condition = f"obj[\"halos\", \"{field}\"]{ustr} {my_op} {value}"
    return condition

def get_about(criterion):
    if len(criterion) == 3:
        field, value, within = criterion
        units = None
    elif len(criterion) == 4:
        field, value, units, within = criterion
    else:
        raise ValueError(
            "Criterion must have either two or three values: "
            f"{criterion}.")

    c1 = get_conditional("above", (field, (1-within)*value, units))
    c2 = get_conditional("below", (field, (1+within)*value, units))
    return [c1, c2]



"""
RockstarArbor class and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
import os
import re

from unyt.exceptions import \
    UnitParseError

from ytree.data_structures.arbor import \
    CatalogArbor
from ytree.frontends.rockstar.fields import \
    RockstarFieldInfo, \
    setup_field_groups
from ytree.frontends.rockstar.io import \
    RockstarDataFile

class RockstarArbor(CatalogArbor):
    """
    Class for Arbors created from Rockstar out_*.list files.
    Use only descendent IDs to determine tree relationship.
    """

    _field_info_class = RockstarFieldInfo
    _data_file_class = RockstarDataFile
    _default_dtype = np.float32

    def _parse_parameter_file(self):
        fgroups = setup_field_groups()
        rems = [f"{s[0]}{t}{s[1]}"
                for s in [("(", ")"), ("", "")]
                for t in ["physical, peculiar",
                          "comoving", "physical"]]

        f = open(self.filename, "r")
        # Read the first line as a list of all fields.
        fields = f.readline()[1:].strip().split()

        # Get box size, cosmological parameters, and units.
        while True:
            line = f.readline()
            if line is None or not line.startswith("#"):
                break
            elif line.startswith("#Om = "):
                pars = line[1:].split(";")
                for j, par in enumerate(["omega_matter",
                                         "omega_lambda",
                                         "hubble_constant"]):
                    v = float(pars[j].split(" = ")[1])
                    setattr(self, par, v)
            elif line.startswith("#Box size:"):
                pars = line.split(":")[1].strip().split()
                self.box_size = self.quan(float(pars[0]), pars[1])
            # Looking for <quantities> in <units>
            elif line.startswith("#Units:"):
                if " in " not in line: continue
                quan, punits = line[8:].strip().split(" in ", 2)
                for rem in rems:
                    while rem in punits:
                        pre, mid, pos = punits.partition(rem)
                        punits = pre + pos
                try:
                    self.quan(1, punits)
                except UnitParseError:
                    punits = ""
                for group in fgroups:
                    if group.in_group(quan):
                        group.units = punits
                        break
        f.close()

        fi = {}
        for i, field in enumerate(fields):
            for group in fgroups:
                units = ""
                if group.in_group(field):
                    units = getattr(group, "units", "")
                    break
            fi[field] = {"column": i, "units": units}

        # the scale factor comes from the catalog file header
        fields.append("scale_factor")
        fi["scale_factor"] = {"source": "header", "units": ""}

        self.field_list = fields
        self.field_info.update(fi)

    def _get_data_files(self):
        """
        Get all out_*.list files and sort them in reverse order.
        """
        reg = re.search(r"_\d+[_\.]", self.basename)
        prefix = self.basename[:reg.start()+1]
        suffix = self.basename[reg.end()-1:]

        freg = re.compile(rf"{prefix}\d+{suffix}")
        my_files = [os.path.join(self.directory, f)
                    for f in os.listdir(self.directory)
                    if freg.match(f)]

        # sort by catalog number
        my_files.sort(
            key=lambda x:
            self._get_file_index(x, prefix, suffix),
            reverse=True)
        self.data_files = \
          [self._data_file_class(f, self) for f in my_files]

    def _get_file_index(self, f, prefix, suffix):
        return int(f[f.find(prefix)+len(prefix):f.rfind(suffix)])

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        File should end in .list.
        """
        fn = args[0]
        if not os.path.basename(fn).startswith("out") or \
          not fn.endswith(".list"):
            return False
        # filter out non-integers
        reg = re.search(r"_\d+[_\.]\D", fn)
        if reg is None:
            return False
        return True



"""
LHaloTreeArbor fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from ytree.data_structures.fields import \
    FieldInfoContainer
from ytree.frontends.lhalotree.utils import \
    dtype_header_default

m_unit = "Msun"
p_unit = "unitary"
r_unit = "kpc"
v_unit = "km/s"

# This is where we set aliases for LHaloTree field names to univeral
# field names.
# Syntax is: (<universal field name>, <field name on disk>, <units>)
# TODO: Use units from parameter file

id_type = np.int64

class LHaloTreeFieldInfo(FieldInfoContainer):
    alias_fields = (
        # ("uid", "uid", None),
        # ("desc_uid", "desc_uid", None),
        # ("scale_factor", "scale_factor", None),
        ("mass", "Mvir", m_unit),
        ("virial_mass", "Mvir", m_unit),
        # ("virial_radius", "Rvir", r_unit),
        # ("scale_radius", "rs", r_unit),
        ("velocity_dispersion", "VelDisp", v_unit),
        ("position_x", "x", p_unit),
        ("position_y", "y", p_unit),
        ("position_z", "z", p_unit),
        ("velocity_x", "vx", v_unit),
        ("velocity_y", "vy", v_unit),
        ("velocity_z", "vz", v_unit),
        ("angular_momentum_x", "Jx", None),
        ("angular_momentum_y", "Jy", None),
        ("angular_momentum_z", "Jz", None),
        # ("spin_parameter", "Spin", None),
    )

    data_types = tuple(
        [("uid", id_type),
         ("desc_uid", id_type)] +
        [(item[0], np.dtype(item[1]))
         for item in dtype_header_default])



"""
loading utilities



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import os

from ytree.config import \
    ytreecfg

if "YTREE_TEST_DATA_DIR" in os.environ:
    test_data_dir = os.environ["YTREE_TEST_DATA_DIR"]
else:
    test_data_dir = ytreecfg["ytree"].get("test_data_dir", ".")

def check_path(filename):
    """
    Check file exists in place or in test data dir.
    """

    if os.path.exists(filename):
        return filename
    tfn = os.path.join(test_data_dir, filename)
    if os.path.exists(tfn):
        return tfn
    raise IOError(f"File does not exist: {filename}.")

def get_path(filename):
    """
    Get a path or list of paths.
    """

    if isinstance(filename, (list, tuple)):
        path = [check_path(fn) for fn in filename]
    else:
        path = check_path(filename)
    return path



"""
TreeFrogArbor io classes and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from collections import defaultdict
import h5py
import numpy as np

from yt.funcs import \
    get_pbar

from ytree.data_structures.io import \
    DataFile, \
    DefaultRootFieldIO, \
    TreeFieldIO
from ytree.utilities.exceptions import \
    ArborFieldNotFound

class TreeFrogDataFile(DataFile):
    def _calculate_arbor_offsets(self):
        """
        Calculate snapshots and snapshot-offsets where each forest appears last.

        For N forests and M snapshots, the data files contain two MxN arrays of:
        - the size of forest i in each snapshot j
        - the offset of forest i in each snapshot j

        In this function, we calculate two things for each forest:
        - the last (i.e. latest in time) snapshot in which the forest exists
        - the file offset within that last snapshot

        Since ytree reads from the root backward, this tell us where to start
        reading for any given forest.
        """
        close = self.fh is None
        self.open()
        fh = self.fh

        offsets = fh["ForestInfoInFile/ForestOffsetsAllSnaps"][()]
        sizes = fh["ForestInfoInFile/ForestSizesAllSnaps"][()]
        size = sizes.shape[0]
        s1 = np.empty(size, dtype=int)
        o1 = np.empty(size, dtype=int)
        for i in range(size):
            s1[i] = np.where(sizes[i] > 0)[0][-1]
            o1[i] = offsets[i, s1[i]]

        if close:
            self.close()

        self._arbor_start = s1
        self._arbor_offset = o1

    def read_data(self, group, field, frange=None):
        if frange is None:
            frange = ()
        # This is the easiest way I can think of to fix the
        # descendent ids of roots.
        if field == "Descendant":
            # Descendant and ID fields are uint64. We need to convert them
            # to signed ints in order to set equal to -1.
            data = self.fh[group][field][()][frange].astype("int64")
            ids = self.fh[group]["ID"][()][frange].astype("int64")
            data[data == ids] = -1
            return data
        return self.fh[group][field][()][frange]

    _arbor_start = None
    @property
    def arbor_start(self):
        if self._arbor_start is None:
            self._calculate_arbor_offsets()
        return self._arbor_start

    _arbor_offset = None
    @property
    def arbor_offset(self):
        if self._arbor_offset is None:
            self._calculate_arbor_offsets()
        return self._arbor_offset

    def open(self):
        if self.fh is None:
            self.fh = h5py.File(self.filename, mode="r")

class TreeFrogTreeFieldIO(TreeFieldIO):
    def _read_fields(self, root_node, fields, dtypes=None,
                     root_only=False):
        """
        Read fields from disk for a single tree.
        """

        if dtypes is None:
            dtypes = {}
        my_dtypes = self._determine_dtypes(
            fields, override_dict=dtypes)

        fi = self.arbor.field_info
        afields = [field for field in fields
                   if fi[field].get("source") == "arbor"]
        rfields = list(set(fields).difference(afields))

        data_file = self.arbor.data_files[root_node._fi]

        close = False
        if data_file.fh is None:
            close = True
            data_file.open()

        fh = data_file.fh
        si = root_node._si

        if not hasattr(root_node, "_offsets"):
            root_node._offsets = fh["ForestInfoInFile/ForestOffsetsAllSnaps"][si]
        offsets = root_node._offsets
        if not hasattr(root_node, "_sizes"):
            root_node._sizes = fh["ForestInfoInFile/ForestSizesAllSnaps"][si]
        sizes = root_node._sizes

        in_snap = np.where(sizes > 0)[0]
        gs = in_snap[-1]
        if root_only:
            ge = gs
        else:
            ge = in_snap[0]

        rdata = defaultdict(list)
        for gi in range(gs, ge-1, -1):
            group = f"Snap_{gi:03d}"
            offset = offsets[gi]
            size = sizes[gi]
            for field in rfields:
                rdata[field].append(
                    data_file.read_data(
                        group, field,
                        frange=slice(offset,offset+size)))

            for field in afields:
                rdata[field].append(self._get_arbor_field(field, gi, size))

        for field, data in rdata.items():
            rdata[field] = np.concatenate(data)

        if close:
            data_file.close()

        field_data = root_node.field_data
        for field in fields:
            field_data[field] = rdata[field]
            dtype = my_dtypes.get(field, fi[field].get("dtype", None))
            if dtype is not None:
                field_data[field] = field_data[field].astype(dtype)

        self._apply_units(fields, field_data)

        return field_data

    def _get_arbor_field(self, field, gi, size):
        if field == "scale_factor":
            return self.arbor._scale_factors[gi] * np.ones(size)
        else:
            raise ArborFieldNotFound(field, arbor=self.arbor)

class TreeFrogRootFieldIO(DefaultRootFieldIO):
    """
    Read fields for the roots of all forests.

    Each data file stores the latest snapshot where a forest exists
    and the offset within that snapshot. We use this information to
    open only the HDF5 groups that have data we need.
    """
    def _read_fields(self, storage_object, fields, dtypes=None):
        if dtypes is None:
            dtypes = {}
        my_dtypes = self._determine_dtypes(
            fields, override_dict=dtypes)

        fi = self.arbor.field_info
        afields = [field for field in fields
                   if fi[field].get("source") == "arbor"]
        rfields = list(set(fields).difference(afields))

        arbor = self.arbor
        arbor._plant_trees()

        # Use the _node_io_loop machinery to get the data files
        # and corresponding tree indices.
        data_files, index_list, _ = arbor._node_io_loop_prepare(None)

        c = 0
        rdata = defaultdict(list)
        pbar = get_pbar('Reading root fields', arbor.size)
        for idf, (data_file, nodes) in enumerate(zip(data_files, index_list)):
            arbor._node_io_loop_start(data_file)

            size = nodes.size
            # the index of the last snapshot for each forest
            s1 = data_file.arbor_start[nodes]
            # the offset within that snapshot
            o1 = data_file.arbor_offset[nodes]

            fdata = {}
            # np.unique(s1) gives us the total number of HDF5 groups we need to open
            for gi in np.unique(s1):
                group = f"Snap_{gi:03d}"
                # which forests' roots are in this group
                isnap = np.where(s1 == gi)[0]

                for field in rfields:
                    gdata = data_file.read_data(group, field)
                    if field not in fdata:
                        fdata[field] = np.empty(size, dtype=gdata.dtype)
                    # o1[isnap] is the list of file offsets for this HDF5 group
                    fdata[field][isnap] = gdata[o1[isnap]]

            for field in rfields:
                rdata[field].append(fdata[field])

            for field in afields:
                rdata[field].append(self._get_arbor_field(field, s1))

            arbor._node_io_loop_finish(data_file)

            c += size
            pbar.update(c)
        pbar.finish()

        field_data = {}
        for field in fields:
            data = np.concatenate(rdata[field])
            dtype = my_dtypes.get(field)
            if dtype is not None:
                data = data.astype(dtype)
            field_data[field] = data

        self._apply_units(fields, field_data)

        return field_data

    def _get_arbor_field(self, field, s1):
        if field == "scale_factor":
            return self.arbor._scale_factors[s1]
        else:
            raise ArborFieldNotFound(field, arbor=self.arbor)



"""
visualization imports



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from functools import wraps
import numpy as np

from unyt import \
    unyt_quantity

try:
    import pydot
except ImportError:
    pydot = None

def clear_graph(f):
    @wraps(f)
    def newfunc(*args, **kwargs):
        rv = f(*args, **kwargs)
        args[0].graph = None
        return rv
    return newfunc

class TreePlot:
    """
    Make a simple merger tree plot using pydot and graphviz.

    Parameters
    ----------
    tree : merger tree node :class:`~ytree.data_structures.tree_node.TreeNode`
        The merger tree to be plotted.
    dot_kwargs : optional, dict
        A dictionary of keyword arguments to be passed to pydot.Dot.
        Default: None.
    node_function: optional, function
        A function accepting a single argument of a
        :class:`~ytree.data_structures.tree_node.TreeNode` and returning a
        dictionary of keywords to be given to pydot for creating the node
        object on the plot. This can be used to customize the appearance of
        the nodes. See examples below.
        Default: None.
    edge_function: optional, function
        A function accepting two
        :class:`~ytree.data_structures.tree_node.TreeNode` objects and
        returning a dictionary of keywords to be given to pydot for creating
        the edge object on the plot (the lines connecting halos). This can
        be used to customize the appearance of the edges. See examples below.
        Default: None.

    Attributes
    ----------
    size_field : str
        The field to determine the size of each circle.
        Default: 'mass'.
    size_log : bool
        Whether to scale circle sizes based on log of size field.
        Default: True.
    min_mass : float or unyt_quantity
        The minimum halo mass to be included in the plot. If given
        as a float, units are assumed to be Msun.
        Default: None.
    min_mass_ratio : float
        The minimum ratio between a halo's mass and the mass of the
        main halo to be included in the plot.
        Default: None.

    Examples
    --------

    >>> import ytree
    >>> a = ytree.load("tree_0_0_0.dat")
    >>> p = ytree.TreePlot(a[0])
    >>> p.min_mass = 1e6 # Msun
    >>> p.save()

    >>> # customizing nodes
    >>> import ytree
    >>> def my_node(halo):
    ...     label = f"{halo['uid']}"
    ...     my_kwargs = {"label": label, "fontsize": 8, "shape": "square"}
    ...     return my_kwargs
    >>> a = ytree.load("tree_0_0_0.dat")
    >>> p = ytree.TreePlot(a[0], node_function=my_node)
    >>> p.save()

    >>> # customizing edges
    >>> import ytree
    >>> def my_edge(ancestor, descendent):
    ...     if descendent['mass'] < ancestor['mass']:
    ...         color = 'blue'
    ...     else:
    ...         color = 'black'
    ...     my_kwargs = {"color": color, "penwidth": 5}
    ...     return my_kwargs
    >>> a = ytree.load("tree_0_0_0.dat")
    >>> p = ytree.TreePlot(a[0], edge_function=my_edge)
    >>> p.save()

    """

    _min_dot_size = 0.2
    _max_dot_size = 2
    _min_field_size = None
    _max_field_size = None

    _size_field = 'mass'
    _size_log = True
    _min_mass = None
    _min_mass_ratio = None

    def __init__(self, tree, dot_kwargs=None,
                 node_function=None, edge_function=None):
        """
        Initialize a TreePlot.
        """

        if pydot is None:
            raise RuntimeError(
                "TreePlot requires the pydot module. " +
                "You may also need to install graphviz.")

        self.tree = tree

        self.dot_kwargs = dict()
        if dot_kwargs is None:
            dot_kwargs = {}
        self.dot_kwargs.update(dot_kwargs)

        if node_function is not None and \
          not callable(node_function):
            raise RuntimeError(
                "node_function should be a callable function.")
        self.node_function = node_function

        if edge_function is not None and \
          not callable(edge_function):
            raise RuntimeError(
                "edge_function should be a callable function.")
        self.edge_function = edge_function

        self.graph = None

    def save(self, filename=None):
        """
        Save the merger tree plot.

        Parameters
        ----------
        filename: optional, str
            The output filename. If none given, the uid of the head
            node is used.
            Default: None.

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("tree_0_0_0.dat")
        >>> p = ytree.TreePlot(a[0])
        >>> p.save('tree.png')

        """

        if filename is None:
            filename = f"tree_{self.tree.uid:06d}.pdf"

        if self.graph is None:
            self._plot()

        suffix = filename[filename.rfind(".")+1:]
        func = getattr(self.graph, f"write_{suffix}", None)
        if func is None:
            raise RuntimeError(f"Cannot save to file format: {suffix}.")

        func(filename)
        return filename

    def _plot(self):
        self.graph = pydot.Dot(graph_type='graph', **self.dot_kwargs)
        self._plot_ancestors(self.tree)

    def _plot_ancestors(self, halo):
        graph = self.graph

        my_node = self._plot_node(halo)
        ancestors = list(halo.ancestors)
        if ancestors is None:
            return

        for anc in ancestors:
            if self.min_mass is not None and \
              anc['mass'] < self.min_mass:
                continue
            if self.min_mass_ratio is not None and \
              anc['mass'] / anc.root['mass'] < self.min_mass_ratio:
                continue

            anc_node = self._plot_node(anc)

            if self.edge_function is not None:
                edge_kwargs = self.edge_function(anc, halo)
            else:
                edge_kwargs = {"penwidth": 5}

            graph.add_edge(pydot.Edge(my_node, anc_node, **edge_kwargs))
            self._plot_ancestors(anc)

    def _plot_node(self, halo):
        graph = self.graph
        node_name = f"{halo.uid}"
        my_node = graph.get_node(node_name)

        if len(my_node) == 0:
            if self.node_function is not None:
                node_kwargs = self.node_function(halo)

            else:
                prog_ids = halo.find_root()['prog', 'uid']
                if halo['uid'] in prog_ids:
                    color = 'red'
                else:
                    color = 'black'
                node_kwargs = \
                  {'style': 'filled', 'label': '', 'fillcolor': color,
                   'shape': 'circle', 'fixedsized': 'true',
                   'width': self._size_norm(halo)}

            my_node = pydot.Node(
                node_name, **node_kwargs)
            graph.add_node(my_node)
        else:
            my_node = my_node[0]

        return my_node

    def _size_norm(self, halo):
        if self._min_field_size is None:
            tdata = self.tree['tree', self.size_field]
            if self.size_log:
                self._min_field_size = tdata[tdata > 0].min()
            else:
                self._min_field_size = tdata.min()
        nmin = self._min_field_size

        if self._max_field_size is None:
            tdata = self.tree['tree', self.size_field]
            self._max_field_size = tdata.max()
        nmax = self._max_field_size

        fval = halo[self.size_field]
        if self.size_log:
            val = np.log(fval / nmin) / np.log(nmax / nmin)
        else:
            val = (fval - nmin) / (nmax - nmin)
        val = np.clip(float(val), 0, 1)

        size = val * (self._max_dot_size - self._min_dot_size) + \
          self._min_dot_size
        return size

    @property
    def min_mass(self):
        """
        The minimum halo mass to be included in the plot.
        """
        return self._min_mass

    @min_mass.setter
    @clear_graph
    def min_mass(self, val):
        if not isinstance(val, unyt_quantity):
            val = unyt_quantity(val, 'Msun')
        self._min_mass = val

    @property
    def min_mass_ratio(self):
        """
        The minimum halo mass to main halo mass.
        """
        return self._min_mass_ratio

    @min_mass_ratio.setter
    @clear_graph
    def min_mass_ratio(self, val):
        self._min_mass_ratio = val

    @property
    def size_field(self):
        """
        The field to determine the size of each circle.
        """
        return self._size_field

    @size_field.setter
    @clear_graph
    def size_field(self, val):
        self._size_field = val

    @property
    def size_log(self):
        """
        Whether to scale circle sizes based on log of size field.
        """
        return self._size_log

    @size_log.setter
    @clear_graph
    def size_log(self, val):
        self._size_log = val



"""
MoriaArbor class and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import h5py
import numpy as np

from yt.funcs import \
    get_pbar

from ytree.data_structures.arbor import \
    Arbor

from ytree.frontends.moria.fields import \
    MoriaFieldInfo
from ytree.frontends.moria.io import \
    MoriaDataFile, \
    MoriaRootFieldIO, \
    MoriaTreeFieldIO
from ytree.utilities.logger import \
    ytreeLogger as mylog

class MoriaArbor(Arbor):
    """
    Arbors from Moria merger trees.
    """

    _suffix = ".hdf5"
    _data_file_class = MoriaDataFile
    _field_info_class = MoriaFieldInfo
    _root_field_io_class = MoriaRootFieldIO
    _tree_field_io_class = MoriaTreeFieldIO
    _node_io_attrs = ('_ai', '_si', '_ei')

    def _parse_parameter_file(self):
        f = h5py.File(self.parameter_filename, mode='r')
        g = f["simulation"]
        self.hubble_constant = float(g.attrs['cosmo_h'])
        self.omega_matter = g.attrs['cosmo_Omega_m']
        self.omega_lambda = g.attrs['cosmo_Omega_L']
        self.omega_radiation = g.attrs['cosmo_Omega_r']
        self.box_size = self.quan(g.attrs['box_size'], 'Mpc/h')
        self._redshifts = self.arr(g.attrs['snap_z'], '')
        self._scale_factors = self.arr(g.attrs['snap_a'], '')
        self._times = self.arr(g.attrs['snap_t'], 'Gyr')

        field_list = []
        fi = {}
        for field in f:
            if isinstance(f[field], h5py.Group):
                continue
            if len(f[field].shape) == 3:
                fields = [f"{field}_{i}" for i in range(f[field].shape[2])]
                field_list.extend(fields)
                fi.update({myf: {"vector": True} for myf in fields})
            else:
                field_list.append(field)
                fi[field] = {}
        f.close()

        field_list.append("snap_index")
        fi["snap_index"] = {"source": "arbor", "dtype": int}

        self.field_list = field_list
        self.field_info.update(fi)

    def _get_data_files(self):
        self.data_files = [self._data_file_class(self.parameter_filename)]

    def _plant_trees(self):
        if self.is_planted:
            return

        f = h5py.File(self.parameter_filename, mode='r')
        status = f["status_sparta"][()]
        root_status = status[-1]
        hosts = root_status == 10
        self._size = hosts.sum()
        self._node_info['_ai'][:] = np.arange(self._size)
        self._node_info['_si'][:] = np.where(hosts)[0]
        self._node_info['_ei'][:-1] = self._node_info['_si'][1:]
        self._node_info['_ei'][-1] = root_status.size
        self._node_info['uid'][:] = f["id"][-1][hosts]
        f.close()

        pbar = get_pbar('Planting trees', self._size)
        si = self._node_info['_si']
        ei = self._node_info['_ei']
        for i in range(self._size):
            tree_status = status[:, si[i]:ei[i]]
            self._node_info['_tree_size'][i] = (tree_status != 0).sum()
            pbar.update(i+1)
        pbar.finish()

    def _setup_tree(self, tree_node, **kwargs):
        """
        Check for desc_uids missing from uid list.
        """

        super()._setup_tree(tree_node, **kwargs)
        uids = tree_node.uids
        desc_uids = tree_node.desc_uids
        missing = np.setdiff1d(desc_uids, uids)
        missing = np.setdiff1d(missing, [-1])
        if missing.size == 0:
            return

        mfields = ["snap_index", "descendant_index"]
        field_data  = self._node_io._read_fields(tree_node, mfields)

        xsize = self._redshifts.size - 1
        ysize = tree_node._ei - tree_node._si
        xindex = xsize - field_data["snap_index"] - 1
        yindex = field_data["descendant_index"] - tree_node._si

        for muid in missing:
            mi = np.where(desc_uids == muid)[0]
            ndi = np.ravel_multi_index([xindex[mi], yindex[mi]], (xsize, ysize))
            for my_mi, my_ndi in zip(mi, ndi):
                new_uid = uids[np.where(my_ndi == tree_node._status)][0]
                mylog.info(f"Reassigning descendent of halo {uids[my_mi]} from "
                           f"{desc_uids[my_mi]} to {new_uid}.")
                desc_uids[my_mi] = new_uid

    def _node_io_loop_prepare(self, nodes):
        self._plant_trees()

        if nodes is None:
            my_size = self.size
        else:
            my_size = nodes.size
        indices = np.arange(my_size)

        return self.data_files, [indices], None

    def _node_io_loop_start(self, data_file):
        data_file.full_read = True

    def _node_io_loop_finish(self, data_file):
        data_file.full_read = False
        data_file.field_cache = None

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        Should be an hdf5 file with a few key attributes.
        """
        fn = args[0]

        if not h5py.is_hdf5(fn):
            return False

        groups = ["config", "simulation"]

        with h5py.File(fn, mode='r') as f:
            for group in groups:
                if group not in f:
                    return False
                if not isinstance(f[group], h5py.Group):
                    return False
        return True



from ytree.frontends.ahf import \
    AHFArbor
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest

class AHFArborTest(TempDirTest, ArborTest):
    arbor_type = AHFArbor
    test_filename = "ahf_halos/snap_N64L16_000.parameter"
    num_data_files = 136
    tree_skip = 100



"""
ConsistentTreesHDF5Arbor io classes and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from collections import defaultdict
import h5py
import numpy as np

from yt.funcs import \
    get_pbar

from ytree.data_structures.io import \
    DataFile, \
    DefaultRootFieldIO, \
    TreeFieldIO

class ChunkStore:
    def __init__(self, chunk_size=262144):
        self.chunk_size = chunk_size
        self.reset()

    def reset(self):
        self.data = {}
        self.ind = {}

    def get(self, fh, field, index):
        start, end = index
        si, ei = self.ind.get(field, (0, 0))

        if field not in self.data or ei < end or si > start:
            si = start
            ei = start + self.chunk_size
            self.ind[field] = (si, ei)
            self.data[field] = fh[field][si:ei]

        data_s = start - si
        data_e = end - si
        return self.data[field][data_s:data_e]

class ConsistentTreesHDF5DataFile(DataFile):
    def __init__(self, filename, linkname):
        super().__init__(filename)
        self.linkname = linkname
        self.real_fh = None
        self._field_cache = ChunkStore()

    def open(self):
        self.real_fh = h5py.File(self.filename, mode="r")
        if self.linkname is None:
            self.fh = self.real_fh
        else:
            self.fh = self.real_fh[self.linkname]

    def close(self):
        self.real_fh.close()
        self.fh = None

class ConsistentTreesHDF5TreeFieldIO(TreeFieldIO):
    def _read_fields(self, root_node, fields, dtypes=None,
                     root_only=False):
        """
        Read fields from disk for a single tree.
        """

        data_file = self.arbor.data_files[root_node._fi]

        close = False
        if data_file.fh is None:
            close = True
            data_file.open()
        fh = data_file.fh['Forests']
        if self.arbor._aos:
            fh = fh['halos']

        if root_only:
            index = (root_node._si, root_node._si+1)
        else:
            index = (root_node._si, root_node._ei)

        field_cache = data_file._field_cache
        field_data = dict((field, field_cache.get(fh, field, index))
                          for field in fields)

        if close:
            data_file.close()
            for field in fields:
                field_data[field] = field_data[field].copy()

        self._apply_units(fields, field_data)

        return field_data

class ConsistentTreesHDF5RootFieldIO(DefaultRootFieldIO):
    """
    Read in fields for first node in all trees/forest.

    This function is optimized for the struct of arrays layout.
    It will work for array of structs layout, but field access
    will be 1 to 2 orders of magnitude slower.
    """
    def _read_fields(self, storage_object, fields, dtypes=None):
        if dtypes is None:
            dtypes = {}

        arbor = self.arbor
        arbor._plant_trees()

        # Use the _node_io_loop machinery to get the data files
        # and corresponding tree indices.
        data_files, index_list, _ = arbor._node_io_loop_prepare(None)

        c = 0
        rdata = defaultdict(list)

        iend = arbor._file_count.cumsum()
        istart = iend - arbor._file_count

        pbar = get_pbar('Reading root fields', arbor.size)
        for idf, (data_file, nodes) in enumerate(zip(data_files, index_list)):
            my_indices = arbor._node_info['_si'][istart[idf]:iend[idf]]
            arbor._node_io_loop_start(data_file)

            fh = data_file.fh['Forests']
            if self.arbor._aos:
                fh = fh['halos']

            for field in fields:
                darray = fh[field][()]
                rdata[field].append(darray[my_indices])

            arbor._node_io_loop_finish(data_file)

            c += my_indices.size
            pbar.update(c)
        pbar.finish()

        field_data = {}
        for field in fields:
            data = np.concatenate(rdata[field])
            dtype = dtypes.get(field)
            if dtype is not None:
                data = data.astype(dtype)
            field_data[field] = data

        self._apply_units(fields, field_data)

        return field_data



"""
ytree io




"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree Development Team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import h5py
from more_itertools import always_iterable
import numpy as np

from yt.utilities.io_handler import BaseIOHandler

_ptype = "halos"

class IOHandlerYTreeHDF5(BaseIOHandler):
    _dataset_type = "ytree_arbor"

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError

    def _read_particle_coords(self, chunks, ptf):
        for data_file in self._yield_data_files(chunks):
            x, y, z = data_file._get_particle_positions(_ptype)
            yield _ptype, (x, y, z), 0.0

    def _read_particle_fields(self, chunks, ptf, selector):
        for data_file in self._yield_data_files(chunks):
            with h5py.File(data_file.filename, "r") as f:
                for ptype, field_list in sorted(ptf.items()):
                    x, y, z = data_file._get_particle_positions(ptype, f=f)
                    mask = selector.select_points(x, y, z, 0.0)
                    del x, y, z
                    if mask is None:
                        continue
                    if mask.all():
                        mask = slice(None)
                    for field in field_list:
                        data = data_file._read_field_data(field, mask, f=f)
                        yield (ptype, field), data

    def _yield_data_files(self, chunks):
        chunks = always_iterable(chunks)
        data_files = set([])
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)

        for data_file in sorted(data_files):
            yield data_file

    def _yield_coordinates(self, data_file):
        pos = data_file._get_particle_positions(_ptype, transpose=False)
        pos.convert_to_units("code_length")
        yield _ptype, pos

    def _count_particles(self, data_file):
        si, ei = data_file.start, data_file.end
        if None not in (si, ei):
            pcount = {}
            for ptype, npart in data_file.total_particles_file.items():
                pcount[ptype] = np.clip(npart - si, 0, ei - si)
        else:
            pcount = data_file.total_particles_file
        return pcount

    def _identify_fields(self, data_file):
        _node_fields = ["file_root_index", "file_number", "tree_index"]
        fields = [(_ptype, field) for field in list(self.ds._field_dict.keys()) +
                  _node_fields]
        units = dict(((_ptype, field), self.ds._field_dict[field].get("units", ""))
                     for field in self.ds._field_dict)
        return fields, units



"""
LHaloTreeHDF5Arbor io classes and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import h5py
import numpy as np
import re

from ytree.data_structures.io import \
    DataFile, \
    TreeFieldIO

class LHaloTreeHDF5DataFile(DataFile):
    def __init__(self, filename, linkname):
        super().__init__(filename)

        self.open()
        self._size = self.fh["Header"].attrs["NtreesPerFile"]
        self.close()

    def open(self):
        self.fh = h5py.File(self.filename, mode="r")

    def close(self):
        self.fh.close()
        self.fh = None

class LHaloTreeHDF5TreeFieldIO(TreeFieldIO):
    def _read_fields(self, root_node, fields, dtypes=None,
                     root_only=False):
        """
        Read fields from disk for a single tree.
        """

        fi = self.arbor.field_info
        afields = [field for field in fields
                   if fi[field].get("source") == "arbor"]
        rfields = list(set(fields).difference(afields))

        for afield in afields:
            rfields.extend(
                [dfield for dfield in fi[afield].get("dependencies", [])
                 if dfield not in rfields])

        data_file = self.arbor.data_files[root_node._fi]
        close = False
        if data_file.fh is None:
            close = True
            data_file.open()
        fh = data_file.fh
        g = fh[f"Tree{root_node._si}"]

        if root_only:
            index = slice(0, 1)
        else:
            index = ()

        field_cache = {}
        field_data = {}
        freg = re.compile(r"(^.+)_(\d+$)")
        for field in rfields:
            fs = freg.search(field)
            if fs and fs.groups()[0] in g:
                fieldname, ifield = fs.groups()
                ifield = int(ifield)
                if fieldname not in field_cache:
                    field_cache[fieldname] = g[fieldname][index]
                field_data[field] = field_cache[fieldname][:, ifield]
            else:
                field_data[field] = g[field][index]

        if afields:
            field_data.update(self._get_arbor_fields(
                root_node, field_data, fields, afields, root_only))

        if close:
            data_file.close()

        self._apply_units(rfields, field_data)

        return field_data

    def _get_arbor_fields(self, root_node, field_data,
                          fields, afields, root_only):
        """
        Generate special fields from the arbor/treenode.
        """

        adata = {}

        if "uid" in afields:
            if root_only:
                adata["uid"] = np.array([root_node.uid])
            else:
                adata["uid"] = root_node.uid + \
                  np.arange(root_node._tree_size)

        if "desc_uid" in afields:
            if "Descendant" in fields:
                desc_uids = field_data["Descendant"].copy()
            else:
                desc_uids = field_data.pop("Descendant")
            desc_uids[desc_uids != -1] += root_node.uid
            adata["desc_uid"] = desc_uids

        return adata



"""
ytree exceptions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

class ArborDataFileEmpty(Exception):
    def __init__(self, filename):
        self.filename = filename

    def __str__(self):
        return (f"Data file is empty: {self.filename}.")

class ArborFieldException(Exception):
    def __init__(self, field, arbor=None):
        self.field = field
        self.arbor = arbor

class ArborDerivedFieldException(ArborFieldException):
    def __str__(self):
        return (f"Derived field \"{self.field}\" could not be "
                f"generated for {self.arbor}.")

class ArborFieldDependencyNotFound(Exception):
    def __init__(self, field, dependency, arbor=None):
        self.field = field
        self.dependency = dependency
        self.arbor = arbor

    def __str__(self):
        return (f"Field dependency not found: \"{self.dependency}\" "
                f"(dependency for \"{self.field}\") in {self.arbor}.")

class ArborFieldCircularDependency(ArborFieldException):
    def __str__(self):
        return f"Field depends on itself: \"{self.field}\" in {self.arbor}."

class ArborFieldNotFound(ArborFieldException):
    def __str__(self):
        return f"Field not found: \"{self.field}\" in {self.arbor}."

class ArborFieldAlreadyExists(ArborFieldException):
    def __str__(self):
        return f"Field already exists: \"{self.field}\" in {self.arbor}."

class ArborAnalysisFieldNotGenerated(ArborFieldException):
    def __str__(self):
        return (f"Analysis field \"{self.field}\" needed but "
                f"not yet generated in {self.arbor}.")

class ArborAnalysisFieldNotFound(ArborFieldException):
    def __str__(self):
        return (f"Analysis field \"{self.field}\" has been removed "
                f"from arbor field storage in {self.arbor}.")

class ArborUnsettableField(ArborFieldException):
    def __str__(self):
        return (f"Cannot set values for field \"{self.field}\" in {self.arbor}. "
                "Only analysis fields can be set.")



"""
consistent_trees_hdf5 frontend



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from ytree.frontends.consistent_trees_hdf5.arbor import \
    ConsistentTreesHDF5Arbor



"""
AHFArbor io classes and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from collections import defaultdict
import numpy as np
import os
import re
import weakref

from ytree.frontends.ahf.misc import \
    parse_AHF_file
from ytree.data_structures.io import \
    CatalogDataFile
from ytree.utilities.io import \
    f_text_block
from ytree.utilities.misc import fround

class AHFDataFile(CatalogDataFile):
    _redshift_precision = 3

    def __init__(self, filename, arbor):
        self.arbor = weakref.proxy(arbor)
        self.filename = filename
        self._catalog_index = arbor._get_file_index(filename)
        self.filekey = self.filename[:self.filename.rfind(self.arbor._par_suffix)]
        self._parse_header()

        rprec = self._redshift_precision
        rstr = r"\.z\d\.\d\{%d\}" % rprec
        res = re.search(rstr, self.filekey)
        if res:
            self.data_filekey = self.filekey[:res.end()]
        else:
            zfmt = f".0{rprec}f"
            for inc in [0, -10**-rprec]:
                my_z = format(fround(self.redshift, decimals=rprec) + inc, zfmt)
                fkey = f"{self.filekey}.z{my_z}"
                if os.path.exists(fkey + self.arbor._data_suffix):
                    self.data_filekey = fkey
                    break

            if not hasattr(self, "data_filekey"):
                raise FileNotFoundError(
                    f"Cannot find data file: {fkey + self.arbor._data_suffix}.")

        self._get_other_filenames()
        self.fh = None
        self._parse_data_header()
        self.offsets = None

    def _get_other_filenames(self):
        """
        Figure out the various additional filenames we're going to need here.

        If prefixes for the ahf and mtree files have been set, then we need to
        look out for this and adjust.
        """

        self.halos_filename = self.data_filekey + self.arbor._data_suffix

        if self.arbor._ahf_prefix is None:
            mtree_key = self.data_filekey
        else:
            # First, strip of the directory
            mtree_key = re.sub(f"^{self.arbor.directory}{os.path.sep}", "",
                               self.data_filekey)
            # Now, substitute the prefixes
            mtree_key = re.sub(f"^{self.arbor._ahf_prefix}",
                               f"{self.arbor._mtree_prefix}",
                               mtree_key)
            # Return the directory
            mtree_key = os.path.join(self.arbor.directory, mtree_key)

        self.mtree_filename = mtree_key + self.arbor._mtree_suffix
        if not os.path.exists(self.mtree_filename):
            self.mtree_filename = None

    def _parse_data_header(self):
        """
        Get header sizes from the two data files ending
        in .AHF_halos and .AHF_mtree.
        """

        self.open()
        fh = self.fh
        fh.seek(0, 2)
        self.file_size = fh.tell()
        fh.seek(0)
        for line, loc in f_text_block(fh):
            if not line.startswith("#"):
                loc -= len(line) + 1
                break
        self._hoffset = loc + len(line) + 1
        self.close()

    def _parse_header(self):
        """
        Get header information from the .log and .parameter
        files. Use that to get the name of the data file.
        """

        vals = parse_AHF_file(
            self.filekey + ".parameter", {"z": "redshift"})

        for par, val in vals.items():
            setattr(self, par, val)

    def open(self):
        if self.fh is None:
            self.fh = open(self.halos_filename, "r")

    _links = None
    @property
    def links(self):
        if self._links is None:
            self._compute_links()
        return self._links

    def _compute_links(self):
        """
        Compute the tree from the graph.

        AHF computes a graph, where a given halo can
        have both multiple progenitors and descendents.

        Use the weight function to determine a unique
        descendent for each halo.

        descendent = max (M_ij = N_ij^2 / (N_i * N_j)),

        where:
           N_ij: number of shared particles
           N_i : number of particles in halo i
           N_j : number of particles in halo j
        """

        data = self._read_mtree()
        if data is None:
            self._links = -1
            return

        m = data["shared"]**2 / (data["prog_part"] * data["desc_part"])

        progids = np.unique(data["prog_id"])
        descids = np.empty(progids.size, dtype=progids.dtype)

        for i, progid in enumerate(progids):
            prf = data["prog_id"] == progid
            descids[i] = data["desc_id"][prf][m[prf].argmax()]
        udata = {"prog_id": progids, "desc_id": descids}

        self._links = udata

    def _read_mtree(self):
        """
        Read map of progenitors to descendents.
        This is the ".AHF_mtree" file.
        """
        if self.mtree_filename is None:
            return None

        data = defaultdict(list)
        descid = descpart = None

        f = open(self.mtree_filename, "r")
        for line, offset in f_text_block(f):
            if line.startswith("#"):
                continue
            if line[0].isdigit():
                oline = line.split()
                descid = int(oline[0])
                descpart = int(oline[1])
            else:
                oline = line.split()
                data["shared"].append(int(oline[0]))
                data["prog_id"].append(int(oline[1]))
                data["prog_part"].append(int(oline[2]))
                data["desc_id"].append(descid)
                data["desc_part"].append(descpart)
        f.close()

        if not data:
            return None

        for field in data:
            data[field] = np.array(data[field])
        return data

    def _read_data_default(self, rfields, dtypes):
        if not rfields:
            return {}

        fi = self.arbor.field_info
        field_data = self._create_field_arrays(rfields, dtypes)
        offsets = []

        self.open()
        f = self.fh
        f.seek(self._hoffset)
        file_size = self.file_size - self._hoffset
        for line, offset in f_text_block(f, file_size=file_size):
            offsets.append(offset)
            sline = line.split()
            for field in rfields:
                dtype = dtypes[field]
                field_data[field].append(dtype(sline[fi[field]["column"]]))
        self.close()

        if self.offsets is None:
            self.offsets = np.array(offsets)

        return field_data

    def _read_data_select(self, rfields, tree_nodes, dtypes):
        if not rfields:
            return {}

        fi = self.arbor.field_info
        nt = len(tree_nodes)
        field_data = self._create_field_arrays(
            rfields, dtypes, size=nt)

        self.open()
        f = self.fh

        for i in range(nt):
            f.seek(self.offsets[tree_nodes[i]._fi])
            line = f.readline()
            sline = line.split()
            for field in rfields:
                dtype = dtypes[field]
                field_data[field][i] = dtype(sline[fi[field]["column"]])
        self.close()

        return field_data

    def _get_mtree_fields(self, tfields, dtypes, field_data):
        """
        Use data from the mtree file to get descendent ids.
        """

        if not tfields:
            return

        links = self.links
        descids = np.empty(
            len(field_data["ID"]),
            dtype=dtypes['desc_id'])

        if self.links == -1:
            descids[:] = -1
        else:
            for i, hid in enumerate(field_data["ID"]):
                inlink = hid == links["prog_id"]
                if not inlink.any():
                    descids[i] = -1
                else:
                    descids[i] = \
                      links["desc_id"][np.where(inlink)[0][0]]

        field_data["desc_id"] = descids

    def _read_fields(self, fields, tree_nodes=None, dtypes=None):
        if dtypes is None:
            dtypes = {}

        fi = self.arbor.field_info

        afields = [field for field in fields
                   if fi[field].get("source") == "arbor"]
        my_fields = set(fields).difference(afields)

        # Separate fields into one that come from the file header,
        # the mtree file, and the halos file.
        data_fields = defaultdict(list)
        for field in my_fields:
            source = fi[field]["file"]
            data_fields[source].append(field)

        hfields = data_fields.pop("header", [])
        rfields = data_fields.pop("halos", [])
        tfields = data_fields.pop("mtree", [])
        # If we needs desc_ids, make sure to get IDs so
        # we can link them.
        if tfields:
            if "ID" not in rfields:
                rfields.append("ID")
                dtypes.update(self.arbor._node_io._determine_dtypes(["ID"]))

        field_data = {}
        if tree_nodes is None:
            field_data.update(
                self._read_data_default(rfields, dtypes))

        else:
            # fields from the actual data
            field_data.update(
                self._read_data_select(rfields, tree_nodes, dtypes))

            # fields from arbor-related info
            field_data.update(
                self._get_arbor_fields(afields, tree_nodes, dtypes))

            # fields from the file header
            field_data.update(
                self._get_header_fields(
                    hfields, tree_nodes, dtypes))

        # use data from the mtree file to get descendent ids
        self._get_mtree_fields(tfields, dtypes, field_data)

        return field_data

class AHFCRMDataFile(AHFDataFile):
    def _compute_links(self):
        """
        Read the CRMratio2 file.
        """

    def _compute_links(self):
        self.arbor._compute_links()

    def _get_mtree_fields(self, tfields, dtypes, field_data):
        if not tfields:
            return

        my_ids = field_data["ID"]
        field_data["desc_id"] = descids = np.full(
            len(my_ids), -1, dtype=dtypes["desc_id"])

        if not self.links:
            return

        for i in range(descids.size):
            descids[i] = self.links.get(my_ids[i], -1)



from ytree.frontends.treefarm import \
    TreeFarmArbor
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest

class TreeFarmArborDescendentsTest(TempDirTest, ArborTest):
    arbor_type = TreeFarmArbor
    test_filename = "tree_farm/tree_farm_descendents/fof_subhalo_tab_000.0.h5"
    num_data_files = 51



"""
tests for yt frontend and selection functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
from numpy.testing import assert_raises
import ytree

from ytree.utilities.testing import \
    TempDirTest, \
    assert_array_rel_equal, \
    requires_file
from ytree.yt_frontend import YTreeDataset

CTG = "tiny_ctrees/locations.dat"

class YTSelectionTest(TempDirTest):
    """
    Test class for selecting halos with yt.
    """

    _setup = False
    def setUp(self):
        super().setUp()
        self._setup = True

    _arbor = None
    @property
    def arbor(self):
        if not self._setup:
            return
        if self._arbor is not None:
            return self._arbor

        a = ytree.load(CTG)
        fn = a.save_arbor()
        self._arbor = ytree.load(fn)
        return self._arbor

    @requires_file(CTG)
    def test_yt_all_data(self):
        a = self.arbor
        ds = a.ytds
        assert isinstance(ds, YTreeDataset)

        ad = ds.all_data()
        for field, units in zip(["mass", "redshift"], ["Msun", ""]):
            yt_data = ad["halos", field].to(units)
            yt_data.sort()
            ytree_data = np.concatenate([t["forest", field] for t in a])
            ytree_data.sort()
            assert_array_rel_equal(yt_data, ytree_data, decimals=5)

    @requires_file(CTG)
    def test_yt_sphere(self):
        a = self.arbor
        ds = a.ytds

        sp = ds.sphere(0.5*ds.domain_center, (20, "Mpc/h"))

        ytree_pos = np.concatenate([t["forest", "position"] for t in a])
        ytree_mass = np.concatenate([t["forest", "mass"] for t in a])
        r = a.quan(sp.radius.to("unitary"))
        c = a.arr(sp.center.to("unitary"))
        ytree_r = np.sqrt(((ytree_pos - c)**2).sum(axis=1))
        in_sphere = ytree_r <= r

        ytree_sp_r = ytree_r[in_sphere].to("unitary")
        ytree_sp_r.sort()
        sp_r = sp["halos", "particle_radius"].to("unitary")
        sp_r.sort()
        assert_array_rel_equal(ytree_sp_r, sp_r, decimals=5)

        sp_mass = sp["halos", "mass"].to("Msun")
        sp_mass.sort()
        ytree_sp_mass = ytree_mass[in_sphere].to("Msun")
        ytree_sp_mass.sort()
        assert_array_rel_equal(ytree_sp_mass, sp_mass, decimals=5)

    @requires_file(CTG)
    def test_above(self):
        a = self.arbor
        ds = a.ytds

        mt = 1e10
        sel = a.get_yt_selection(above=[("mass", mt, "Msun")])
        sel_mass = sel["halos", "mass"].to("Msun")
        assert (sel_mass >= mt).all()
        sel_mass.sort()

        ad = ds.all_data()
        ad_mass = ad["halos", "mass"].to("Msun")
        yt_mass = ad_mass[ad_mass >= mt]
        yt_mass.sort()
        assert_array_rel_equal(sel_mass, yt_mass, decimals=5)

    @requires_file(CTG)
    def test_above_no_units(self):
        a = self.arbor
        ds = a.ytds

        mt = 1e10
        sel = a.get_yt_selection(above=[("mass", mt)])
        sel_mass = sel["halos", "mass"].to("Msun")
        assert (sel_mass >= mt).all()
        sel_mass.sort()

        ad = ds.all_data()
        ad_mass = ad["halos", "mass"].to("Msun")
        yt_mass = ad_mass[ad_mass >= mt]
        yt_mass.sort()
        assert_array_rel_equal(sel_mass, yt_mass, decimals=5)

    @requires_file(CTG)
    def test_above_sphere(self):
        a = self.arbor
        ds = a.ytds
        sp = ds.sphere(0.5*ds.domain_center, (20, "Mpc/h"))

        mt = 1e10
        sel = a.get_yt_selection(above=[("mass", mt, "Msun")],
                                 data_source=sp)
        sel_mass = sel["halos", "mass"].to("Msun")
        assert (sel_mass >= mt).all()
        sel_mass.sort()

        sp_mass = sp["halos", "mass"].to("Msun")
        yt_mass = sp_mass[sp_mass >= mt]
        yt_mass.sort()
        assert_array_rel_equal(sel_mass, yt_mass, decimals=5)

    @requires_file(CTG)
    def test_below(self):
        a = self.arbor
        ds = a.ytds

        mt = 1e10
        sel = a.get_yt_selection(below=[("mass", mt, "Msun")])
        sel_mass = sel["halos", "mass"].to("Msun")
        assert (sel_mass <= mt).all()
        sel_mass.sort()

        ad = ds.all_data()
        ad_mass = ad["halos", "mass"].to("Msun")
        yt_mass = ad_mass[ad_mass <= mt]
        yt_mass.sort()
        assert_array_rel_equal(sel_mass, yt_mass, decimals=5)

    @requires_file(CTG)
    def test_below_no_units(self):
        a = self.arbor
        ds = a.ytds

        mt = 1e10
        sel = a.get_yt_selection(below=[("mass", mt)])
        sel_mass = sel["halos", "mass"].to("Msun")
        assert (sel_mass <= mt).all()
        sel_mass.sort()

        ad = ds.all_data()
        ad_mass = ad["halos", "mass"].to("Msun")
        yt_mass = ad_mass[ad_mass <= mt]
        yt_mass.sort()
        assert_array_rel_equal(sel_mass, yt_mass, decimals=5)

    @requires_file(CTG)
    def test_below_sphere(self):
        a = self.arbor
        ds = a.ytds
        sp = ds.sphere(0.5*ds.domain_center, (20, "Mpc/h"))

        mt = 1e10
        sel = a.get_yt_selection(below=[("mass", mt, "Msun")],
                                 data_source=sp)
        sel_mass = sel["halos", "mass"].to("Msun")
        assert (sel_mass <= mt).all()
        sel_mass.sort()

        sp_mass = sp["halos", "mass"].to("Msun")
        yt_mass = sp_mass[sp_mass <= mt]
        yt_mass.sort()
        assert_array_rel_equal(sel_mass, yt_mass, decimals=5)

    @requires_file(CTG)
    def test_about(self):
        a = self.arbor
        ds = a.ytds

        mt = 1e10
        within = 0.5
        sel = a.get_yt_selection(about=[("mass", mt, "Msun", within)])
        sel_mass = sel["halos", "mass"].to("Msun")
        assert ((sel_mass >= (1-within)*mt) & (sel_mass <= (1+within)*mt)).all()
        sel_mass.sort()

        ad = ds.all_data()
        ad_mass = ad["halos", "mass"].to("Msun")
        yt_mass = ad_mass[(ad_mass >= (1-within)*mt) & (ad_mass <= (1+within)*mt)]
        yt_mass.sort()
        assert_array_rel_equal(sel_mass, yt_mass, decimals=5)

    @requires_file(CTG)
    def test_about_no_units(self):
        a = self.arbor
        ds = a.ytds

        mt = 1e10
        within = 0.5
        sel = a.get_yt_selection(about=[("mass", mt, within)])
        sel_mass = sel["halos", "mass"].to("Msun")
        assert ((sel_mass >= (1-within)*mt) & (sel_mass <= (1+within)*mt)).all()
        sel_mass.sort()

        ad = ds.all_data()
        ad_mass = ad["halos", "mass"].to("Msun")
        yt_mass = ad_mass[(ad_mass >= (1-within)*mt) & (ad_mass <= (1+within)*mt)]
        yt_mass.sort()
        assert_array_rel_equal(sel_mass, yt_mass, decimals=5)

    @requires_file(CTG)
    def test_about_sphere(self):
        a = self.arbor
        ds = a.ytds
        sp = ds.sphere(0.5*ds.domain_center, (20, "Mpc/h"))

        mt = 1e10
        within = 0.5
        sel = a.get_yt_selection(about=[("mass", mt, within)],
                                 data_source=sp)
        sel_mass = sel["halos", "mass"].to("Msun")
        assert ((sel_mass >= (1-within)*mt) & (sel_mass <= (1+within)*mt)).all()
        sel_mass.sort()

        sp_mass = sp["halos", "mass"].to("Msun")
        yt_mass = sp_mass[(sp_mass >= (1-within)*mt) & (sp_mass <= (1+within)*mt)]
        yt_mass.sort()
        assert_array_rel_equal(sel_mass, yt_mass, decimals=5)

    @requires_file(CTG)
    def test_equal(self):
        a = self.arbor
        ds = a.ytds

        sel = a.get_yt_selection(equal=[("mmp?", 1, "")])
        sel_mmp = sel["halos", "mmp?"].to("")
        assert (sel_mmp == 1).all()
        sel_mmp.sort()

        ad = ds.all_data()
        ad_mmp = ad["halos", "mmp?"].to("")
        yt_mmp = ad_mmp[ad_mmp == 1]
        yt_mmp.sort()
        assert_array_rel_equal(sel_mmp, yt_mmp, decimals=5)

    @requires_file(CTG)
    def test_equal_no_units(self):
        a = self.arbor
        ds = a.ytds

        sel = a.get_yt_selection(equal=[("mmp?", 1)])
        sel_mmp = sel["halos", "mmp?"].to("")
        assert (sel_mmp == 1).all()
        sel_mmp.sort()

        ad = ds.all_data()
        ad_mmp = ad["halos", "mmp?"].to("")
        yt_mmp = ad_mmp[ad_mmp == 1]
        yt_mmp.sort()
        assert_array_rel_equal(sel_mmp, yt_mmp, decimals=5)

    @requires_file(CTG)
    def test_equal_sphere(self):
        a = self.arbor
        ds = a.ytds
        sp = ds.sphere(0.5*ds.domain_center, (20, "Mpc/h"))

        sel = a.get_yt_selection(equal=[("mmp?", 1, "")],
                                 data_source=sp)
        sel_mmp = sel["halos", "mmp?"].to("")
        assert (sel_mmp == 1).all()
        sel_mmp.sort()

        sp_mmp = sp["halos", "mmp?"].to("")
        yt_mmp = sp_mmp[sp_mmp == 1]
        yt_mmp.sort()
        assert_array_rel_equal(sel_mmp, yt_mmp, decimals=5)

    @requires_file(CTG)
    def test_conditionals(self):
        a = self.arbor
        ds = a.ytds

        sel = a.get_yt_selection(conditionals=['obj["halos", "mass"] > 1e10'])
        sel_mass = sel["halos", "mass"].to("Msun")
        assert (sel_mass >= 1e10).all()
        sel_mass.sort()

        ad = ds.all_data()
        ad_mass = ad["halos", "mass"].to("Msun")
        yt_mass = ad_mass[ad_mass >= 1e10]
        yt_mass.sort()
        assert_array_rel_equal(sel_mass, yt_mass, decimals=5)

    @requires_file(CTG)
    def test_conditionals_sphere(self):
        a = self.arbor
        ds = a.ytds
        sp = ds.sphere(0.5*ds.domain_center, (20, "Mpc/h"))

        sel = a.get_yt_selection(conditionals=['obj["halos", "mass"] > 1e10'],
                                 data_source=sp)
        sel_mass = sel["halos", "mass"].to("Msun")
        assert (sel_mass >= 1e10).all()
        sel_mass.sort()

        sp_mass = sp["halos", "mass"].to("Msun")
        yt_mass = sp_mass[sp_mass >= 1e10]
        yt_mass.sort()
        assert_array_rel_equal(sel_mass, yt_mass, decimals=5)

    @requires_file(CTG)
    def test_yt_selection_bad_input(self):
        a = self.arbor

        with assert_raises(ValueError):
            a.get_yt_selection(
                conditionals=['obj["halos", "mass"] > 1e10'],
                above=["mass", 1e10])

    @requires_file(CTG)
    def test_multiple_criteria_1(self):
        a = self.arbor
        ds = a.ytds

        sel = a.get_yt_selection(above=[("mass", 1e10, "Msun"),
                                        ("redshift", 0.5)])
        sel_mass = sel["halos", "mass"].to("Msun")
        assert (sel_mass >= 1e10).all()
        sel_redshift = sel["halos", "redshift"]
        assert (sel_redshift >= 0.5).all()
        sel_mass.sort()

        ad = ds.all_data()
        ad_mass = ad["halos", "mass"].to("Msun")
        ad_redshift = ad["halos", "redshift"]
        yt_mass = ad_mass[(ad_mass >= 1e10) & (ad_redshift >= 0.5)]
        yt_mass.sort()
        assert_array_rel_equal(sel_mass, yt_mass, decimals=5)

    @requires_file(CTG)
    def test_multiple_criteria_2(self):
        a = self.arbor
        ds = a.ytds

        sel = a.get_yt_selection(above=[("mass", 1e10, "Msun")],
                                 below=[("redshift", 0.5)])
        sel_mass = sel["halos", "mass"].to("Msun")
        assert (sel_mass >= 1e10).all()
        sel_redshift = sel["halos", "redshift"]
        assert (sel_redshift <= 0.5).all()
        sel_mass.sort()

        ad = ds.all_data()
        ad_mass = ad["halos", "mass"].to("Msun")
        ad_redshift = ad["halos", "redshift"]
        yt_mass = ad_mass[(ad_mass >= 1e10) & (ad_redshift <= 0.5)]
        yt_mass.sort()
        assert_array_rel_equal(sel_mass, yt_mass, decimals=5)

    @requires_file(CTG)
    def test_nodes_from_sphere(self):
        a = self.arbor
        ds = a.ytds
        sp = ds.sphere(0.5*ds.domain_center, (20, "Mpc/h"))

        nodes = list(a.get_nodes_from_selection(sp))

        sp_mass = sp["halos", "mass"].to("Msun")
        sp_mass.sort()
        node_mass = a.arr([node["mass"] for node in nodes])
        node_mass.sort()
        assert_array_rel_equal(node_mass, sp_mass, decimals=5)

        node_pos = a.arr([node["position"] for node in nodes])
        r = a.quan(sp.radius.to("unitary"))
        c = a.arr(sp.center.to("unitary"))
        node_r = np.sqrt(((node_pos - c)**2).sum(axis=1))
        assert (node_r <= r).all()

    @requires_file(CTG)
    def test_nodes_from_selection(self):
        a = self.arbor

        sel = a.get_yt_selection(above=[("mass", 1e10, "Msun")])
        sel_mass = sel["halos", "mass"].to("Msun")
        sel_mass.sort()

        nodes = list(a.get_nodes_from_selection(sel))
        node_mass = a.arr([node["mass"] for node in nodes]).to("Msun")
        node_mass.sort()

        assert_array_rel_equal(node_mass, sel_mass, decimals=5)



"""
YTreeArbor io classes and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import h5py
import numpy as np

from ytree.data_structures.io import \
    DefaultRootFieldIO, \
    DataFile, \
    TreeFieldIO

class YTreeDataFile(DataFile):
    def __init__(self, filename):
        super().__init__(filename)
        self._field_cache = None
        self._start_index = None
        self._end_index = None

    def open(self):
        self.fh = h5py.File(self.filename, mode="r")
        if hasattr(self, "analysis_filename"):
            self.analysis_fh = h5py.File(self.analysis_filename, mode="r")

    def close(self):
        self.fh.close()
        self.fh = None
        if hasattr(self, "analysis_filename"):
            self.analysis_fh.close()
            self.analysis_fh = None

class YTreeTreeFieldIO(TreeFieldIO):
    def _read_fields(self, root_node, fields, dtypes=None, root_only=False):
        dfi = np.digitize(root_node._ai, self._ei)
        data_file = self.arbor.data_files[dfi]

        if dtypes is None:
            dtypes = {}

        if data_file._field_cache is None:
            data_file._field_cache = {}

        if data_file.fh is None:
            close = True
            data_file.open()
        else:
            close = False

        # get start_index and end_index
        for itype in ["start", "end"]:
            if getattr(data_file, f"_{itype}_index") is None:
                setattr(data_file, f"_{itype}_index",
                        data_file.fh[f"index/tree_{itype}_index"][()])
        ii = root_node._ai - self._si[dfi]

        field_data = {}
        fi = self.arbor.field_info
        for field in fields:
            if field not in data_file._field_cache:
                if fi[field].get("type") == "analysis_saved":
                    fh = data_file.analysis_fh
                else:
                    fh = data_file.fh
                fdata = fh[f"data/{field}"][()]

                dtype = dtypes.get(field)
                if dtype is not None:
                    fdata = fdata.astype(dtype)

                units = fi[field].get("units", "")
                if units != "":
                    fdata = self.arbor.arr(fdata, units)
                data_file._field_cache[field] = fdata

            field_data[field] = \
              data_file._field_cache[field][
                  data_file._start_index[ii]:data_file._end_index[ii]]

        if close:
            data_file.close()

        return field_data

class YTreeRootFieldIO(DefaultRootFieldIO):
    def _read_fields(self, storage_object, fields, dtypes=None):
        if dtypes is None:
            dtypes = {}

        fh = h5py.File(self.arbor.filename, mode="r")
        if self.arbor.analysis_filename is not None:
            analysis_fh = h5py.File(self.arbor.analysis_filename, mode="r")

        field_data = {}
        fi = self.arbor.field_info
        for field in fields:
            if fi[field].get("type") == "analysis_saved":
                my_fh = analysis_fh
            else:
                my_fh = fh
            data = my_fh[f"data/{field}"][()]
            dtype = dtypes.get(field)
            if dtype is not None:
                data = data.astype(dtype)
            field_data[field] = data

        self._apply_units(fields, field_data)

        fh.close()
        if self.arbor.analysis_filename is not None:
            analysis_fh.close()

        return field_data



"""
LHaloTreeArbor class and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
import glob

from yt.funcs import \
    get_pbar
from unyt.exceptions import \
    UnitParseError

from ytree.data_structures.arbor import \
    Arbor

from ytree.frontends.lhalotree.fields import \
    LHaloTreeFieldInfo
from ytree.frontends.lhalotree.io import \
    LHaloTreeTreeFieldIO, LHaloTreeRootFieldIO
from ytree.frontends.lhalotree.utils import \
    LHaloTreeReader
from ytree.utilities.logger import \
    ytreeLogger


class LHaloTreeArbor(Arbor):
    """
    Arbors for LHaloTree data.
    """

    _field_info_class = LHaloTreeFieldInfo
    _tree_field_io_class = LHaloTreeTreeFieldIO
    _root_field_io_class = LHaloTreeRootFieldIO
    _default_dtype = np.float32
    _node_io_attrs = ('_lht', '_index_in_lht')

    def __init__(self, *args, **kwargs):
        r"""Added reader class to allow fast access of header info."""
        reader_keys = ['parameters', 'parameter_file',
                       'scale_factors', 'scale_factor_file',
                       'header_size', 'nhalos_per_tree', 'read_header_func',
                       'item_dtype']
        reader_kwargs = dict()
        for k in reader_keys:
            if k in kwargs:
                reader_kwargs[k] = kwargs.pop(k)
        self._lht0 = LHaloTreeReader(args[0], **reader_kwargs)
        super().__init__(*args, **kwargs)
        kwargs.update(**reader_kwargs)
        lht0 = self._lht0
        files = sorted(glob.glob(lht0.filepattern))
        self._lhtfiles = [None for _ in files]
        self._lhtfiles[lht0.fileindex] = lht0
        if len(files) > 1:
            if ((('header_size' in reader_kwargs) or  # pragma: no cover
                 ('nhalos_per_tree' in reader_kwargs))):
                raise RuntimeError("Cannot use 'header_size' or 'nhalos_per_tree' " +
                                   "for trees split across multiple files. Use " +
                                   "'read_header_func' instead.")
            reader_kwargs.update(parameters=lht0.parameters,
                                 scale_factors=lht0.scale_factors,
                                 item_dtype=lht0.item_dtype,
                                 silent=True)
            for f in files:
                if f == lht0.filename:
                    continue
                ilht = LHaloTreeReader(f, **reader_kwargs)
                self._lhtfiles[ilht.fileindex] = ilht
        # Assert files are there
        for f in self._lhtfiles:
            if f is None:  # pragma: no cover
                raise RuntimeError("Not all files were read.")

    # NOTE: LHaloTree is currently using np.memmap so it dosn't need
    # fd to be passed. If this causes memory issues for larger trees,
    # the below functions allow a file descriptor to be checked and updated
    # as needed to prevent unnecessary open/close operations when accessing
    # nodes in the same file. Alternatively, see io.py for an option that
    # caches the file descriptor any time a field is read.

    # def _func_update_file(self, node, *args, **kwargs):
    #     """Call a file making sure that the correct file is open."""
    #     func = kwargs.pop("_func", None)
    #     if func is None:
    #         raise RuntimeError("No function passed.")
    #     fd = self._node_io_fd
    #     if fd is None or (node._lht.filename != fd.name):
    #         if fd is not None:
    #             fd.close()
    #         self._node_io_fd = open(node._lht.filename, 'rb')
    #     kwargs["f"] = self._node_io_fd
    #     return func(node, *args, **kwargs)

    # def _node_io_loop(self, func, *args, **kwargs):
    #     """
    #     Since LHaloTrees can be split across multiple files, this
    #     optimization only works if the nodes are all in the same file.
    #     It's a small optimization that keeps the file open
    #     when doing io for multiple trees in the same file.
    #     """
    #     self._node_io_fd = None
    #     kwargs["_func"] = func
    #     super()._node_io_loop(
    #         self._func_update_file, *args, **kwargs)
    #     if self._node_io_fd is not None:
    #         self._node_io_fd.close()

    def _parse_parameter_file(self):
        """
        Parse the file header, get things like:
        - cosmological parameters
        - box size
        - list of fields
        """

        for u in ['mass', 'vel', 'len']:
            setattr(self, '_lht_units_' + u,
                    getattr(self._lht0, 'units_' + u))
            # v, s = getattr(self._lht0, 'units_' + u).split()
            # setattr(self, '_lht_units_' + u, self.quan(float(v), s))

        self.hubble_constant = self._lht0.hubble_constant
        self.omega_matter = self._lht0.omega_matter
        self.omega_lambda = self._lht0.omega_lambda
        self.box_size = self.quan(self._lht0.box_size, self._lht_units_len)
        # self.box_size = self._lht0.box_size * self._lht_units_len

        # a list of all fields on disk
        fields = self._lht0.fields
        # a dictionary of information for each field
        # this can have specialized information for reading the field
        fi = {}
        # for example:
        # fi["mass"] = {"column": 4, "units": "Msun/h", ...}
        none_keys = ['Descendant', 'FirstProgenitor', 'NextProgenitor',
                     'FirstHaloInFOFgroup', 'NextHaloInFOFgroup',
                     'Len', 'MostBoundID',
                     'SnapNum', 'FileNr', 'SubhaloIndex',
                     'uid', 'desc_uid', 'scale_factor',
                     'Jx', 'Jy', 'Jz']
        mass_keys = ['M_Mean200', 'Mvir', 'M_TopHat', 'SubHalfMass']
        dist_keys = ['x', 'y', 'z']
        velo_keys = ['VelDisp', 'Vmax', 'vx', 'vy', 'vz']
        all_keys = [none_keys, mass_keys, dist_keys, velo_keys]
        all_units = ['', self._lht_units_mass, self._lht_units_len,
                     self._lht_units_vel]
        for keylist, unit in zip(all_keys, all_units):
            try:
                self.quan(1, unit)
                punit = unit
            except UnitParseError:  # pragma: no cover
                ytreeLogger.warning(f"Could not parse unit: {unit}")
                punit = ''
            for k in keylist:
                fi[k] = {'units': punit}
            
        self.field_list = fields
        self.field_info.update(fi)

    def _plant_trees(self):
        """
        This is where we figure out how many trees there are,
        create the array to hold them, and instantiate a root
        tree_node for each tree.
        """

        if self.is_planted:
            return

        # open file, count trees
        ntrees_tot = 0
        for lht in self._lhtfiles:
            ntrees_tot += lht.ntrees
        self._size = ntrees_tot

        pbar = get_pbar("Loading tree roots", ntrees_tot)
        self._node_info['_lht'] = np.empty(ntrees_tot, dtype=object)

        itot = 0
        for ifile, lht in enumerate(self._lhtfiles):
            ntrees = lht.ntrees
            root_uids = lht.all_uids[lht.nhalos_before_tree]
            for i in range(ntrees):
                self._node_info['uid'][itot] = root_uids[i]
                self._node_info['_lht'][itot] = lht
                self._node_info['_index_in_lht'][itot] = i
                itot += 1
                pbar.update(itot)

        pbar.finish()

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        Return True if we are able to initialize a reader.
        """
        try:
            kwargs.update(silent=True, validate=True)
            LHaloTreeReader(*args, **kwargs)
        except (IOError, TypeError):
            return False
        return True
